# Copyright 2016-2018 Sebastian Alarcon Villaseca, Sebastián Alarcón Villaseca, Fawzi Mohamed, Micael Oliveira, Ankit Kariryaa, Danio Brambila
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# -*- coding: utf-8 -*
from builtins import object
import numpy as np
from nomadcore.simple_parser import SimpleMatcher, mainFunction
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import os, sys, json, logging
SM=SimpleMatcher

##########################################################
#							 #
#		Parser for ORCA output file		 #
#							 #
##########################################################

##########################################################
###############[1] transfer PARSER CONTEXT ###############
##########################################################

logger = logging.getLogger("nomad.orcaParser")

class OrcaContext(object):
    """context for the sample parser"""

    def __init__(self):
        self.parser = None
        self.CalculationGIndex = None

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse different files"""
        self.CalculationGIndex = None

    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser

        # save metadata
        self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv

        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    def onClose_x_orca_atom_positions(self, backend, gIndex, value):
            x = value["x_orca_atom_positions_x"]
       	    y = value["x_orca_atom_positions_y"]
       	    z = value["x_orca_atom_positions_z"]
            pos = np.zeros((len(x),3), dtype=float)
            pos[:, 0] = x
       	    pos[:, 1] = y
       	    pos[:, 2] = z
            backend .addArrayValues("atom_positions", pos)
            backend.addValue("atom_labels", value["x_orca_atom_labels"])

    def onClose_x_orca_final_geometry(self, backend, gIndex, value):
            x = value["x_orca_atom_positions_x_geo_opt"]
            y = value["x_orca_atom_positions_y_geo_opt"]
            z = value["x_orca_atom_positions_z_geo_opt"]
            pos = np.zeros((len(x),3), dtype=float)
            pos[:,0] = x
            pos[:,1] = y
            pos[:,2] = z
            backend.addArrayValues("geometry_optimization_converged", pos)
            backend.addValue("atom_labels", value["x_orca_atom_labels_geo_opt"])

    def onClose_x_orca_section_functionals(self, backend, gIndex, section):
        functional_names = section["x_orca_XC_functional_type"]

        if functional_names == None: functional = "HF" #default method is Hartree-Fock
        else: functional = functional_names[-1]

        if functional:
            functionalMap = {

		#***************************************
		# Local functionals
		#***************************************
		# HFS: Hartree-Fock Slater
                "HF":     ["HF_X"],
                "HFS":    ["HF_X"],
		# XAlpha: The famous old Slater Xa theory
		"XAlpha": ["HF_X", "LDA_C_XALPHA"],
		# LSD: Local spin density (VWN-5A form)
		"LSD":    ["LDA_X", "LDA_C_VWN_RPA", "VWN5_RPA"],
                "LDA":    ["LDA_X", "LDA_C_VWN_RPA", "VWN5_RPA"],
		# VWN5: Local spin density (VWN-5)
		"VWN":    ["LDA_X", "LDA_C_VWN_RPA", "VWN5_RPA"],
                "VWN5":   ["LDA_X", "LDA_C_VWN_RPA", "VWN5_RPA"],
		# VWN3: Local spin density (VWN-3)
		"VWN3":   ["LDA_X", "LDA_C_VWN_3"],
		# PWLDA: Local spin density (PW-LDA)
		"PWLDA":  ["LDA_X", "LDA_C_PW"],
		#***************************************
		# ’Pure’ GGA functionals
		#***************************************
		# BNULL: Becke ’88 exchange, no corr.
		"BNULL":  ["GGA_X_B88"],
		# BVWN: Becke ’88 exchane, VWN-5 corr.
		"BVWN":   ["GGA_X_B88", "LDA_C_VWN_RPA", "VWN5_RPA"],
		# BP: Becke X-Perdew 86 correlation
		"BP":     ["LDA_X", "GGA_X_B88", "LDA_C_VWN", "GGA_C_P86"],
		"BP86":   ["LDA_X", "GGA_X_B88", "LDA_C_VWN", "GGA_C_P86"],
		# PW91: Perdew-Wang GGA-II ’91 func.
		"PW91":   ["GGA_X_PW91", "GGA_C_PW91"],
		# PWP: Perdew-Wang ’91 exchange and Perdew ’86 correlation
                "PWP":    ["GGA_X_PW91", "GGA_C_P86"],
		# mPWPW: Modified PW with PW correlation
		"mPWPW":  ["GGA_X_MPW91", "GGA_C_PW91"],
		# mPWLYP: same with LYP correlation
		"mPWLYP": ["GGA_X_MPW91", "GGA_C_LYP"],
		# BLYP: Becke X with LYP correlation
		"BLYP":   ["GGA_X_B88", "GGA_C_LYP"],
		# GP: Gill ’96 X, Perdew ’86 corr.
		"GP":     ["GGA_X_G96", "GGA_C_P86"],
		#GLYP: Gill ’96 X with LYP correlation
		"GLYP":   ["GGA_X_G96", "GGA_C_LYP"],
		# PBE: Perdew-Burke-Ernzerhof
		"PBE":    ["GGA_X_PBE", "GGA_C_PBE"],
		# revPBE: Revised PBE (exchange scaling)
		"revPBE": ["GGA_X_PBE_R", "GGA_C_PBE"],
		# RPBE: Revised PBE (functional form of X)
		"RPBE":   ["GGA_X_RPBE", "GGA_C_PBE"],
		# PWP: PW91 exchange + P86 correlation
		"PW91":   ["GGA_X_PW91", "GGA_C_P86"],
		# OLYP: Handy’s ‘optimal’ exchange and Lee-Yang-Parr correlation (the optimized exchange and LYP)
		"OLYP":   ["GGA_X_OPTX", "GGA_C_LYP", "GGA_XC_OPWLYP_D"],
		# OPBE: the optimized exchange and PBE
		"OPBE":   ["GGA_XC_OPBE_D"],
		#XLYP: the Xu/Goddard exchange and LYP
		"XLYP":   ["GGA_XC_XLYP"],
		# B97-D: Grimme’s GGA including D2 dispersion correction
		"B97-D":  ["GGA_XC_B97_D"],
                "B97-D3": ["GGA_XC_B97_D"],
		#***************************************
		# Meta GGA and hybrid meta-GGA functals.
		#***************************************
		# TPSS: the TPPS functional
		"TPSS":   ["LDA_X", "MGGA_X_TPSS", "LDA_C_PW", "MGGA_C_TPSS"],
		# TPSSh: The hybrid version of TPSS (10% HF exchange)
		"TPSSh":  ["HYB_MGGA_XC_TPSSH"],
		# TPSS0: A 25% exchange version of TPSSh that yields improved energetics
                # compared to TPSSh but is otherwise not well tested
		# "TPSS0":  ["HYB_MGGA_XC_TPSSH"],
		# M06L: The Minnesota M06-L meta-GGA functional
		"M06L":   ["MGGA_X_M06_L", "MGGA_C_M06_L"],
		# M06: The M06 hybrid meta-GGA (27% HF exchange)
		"M06":    ["MGGA_X_M06", "MGGA_C_M06"],
		# M062X: The M06-2X version with 54% HF exchange
		"M062X":  ["MGGA_X_M06_2X", "MGGA_C_M06_2X"],
		#***************************************
		# Hybrid functionals
		#***************************************
		# B1LYP: One parameter Hybrid of BLYP
		"B1LYP":     ["HYB_GGA_XC_B1LYP"],
		# B3LYP: Three parameter Hybrid of BLYP
                "B3LYP":     ["HYB_GGA_XC_B3LYP"],
		# B1P: Analogous with Perdew exchange (The one-parameter hybrid version of BP86)
		"B1P":       ["HYB_GGA_XC_B1PW91"],
		# B3P: Analogous with Perdew exchange (The three-parameter hybrid version of BP86)
		"B3P":       ["HYB_GGA_XC_B3P86"],
		# B3PW: The three-parameter hybrid version of PW91
		"B3PW":      ["HYB_GGA_XC_B3PW91"],
		# G1LYP: 1 par. analog with Gill 96 X
		"G1LYP":     ["GGA_X_G96", "GGA_C_OP_G96"],
		# G3LYP: 3 par. analog with Gill 96 X
		# G1P: similar with P correlation
		# G3P # similar with P correlation
		# PBE0: 1 parameter version of PBE (PBEH, PBE0)
		"PBE0":      ["HYB_GGA_XC_PBEH"],
		# PW1PW: One-parameter hybrid version of PW91
		"PW1PW":     ["HYB_GGA_XC_B1PW91"],
                "PWP91_1":   ["HYB_GGA_XC_B1PW91"],
		"PWP1":      ["HYB_GGA_XC_B1PW91"],
		# mPW1PW: 1 parameter version of mPWPW
		"mPW1PW":    ["HYB_GGA_XC_MPW1PW"],
		# mPW1LYP: One-parameter hybrid version of mPWLYP
		"mPW1LYP":   ["HYB_GGA_XC_MPWLYP1M"],
		# O3LYP: 3 parameter version of OLYP
		"O3LYP":     ["HYB_GGA_XC_O3LYP"],
		# X3LYP: 3 parameter version of XLYP
		"X3LYP":     ["HYB_GGA_XC_X3LYP"],
		# PW6B95: Hybrid functional by Truhlar [93]
		"PW6B95":    ["HYB_MGGA_XC_PW6B95"],
		# B97: Becke’s original hybrid
		"B97":       ["HYB_GGA_XC_B97"],
		# BHANDHLYP: Half-and-half Becke hybrid functional
		"BHANDHLYP": ["HYB_GGA_XC_BHANDHLYP"],
		#***************************************
		# Double Hybrid functionals (mix in MP2)
		#***************************************
		# B2PLYP: Grimme’s 2006 double hybrid
		"B2PLYP":    ["HYB_GGA_XC_B2PLYP"],
		# mPW2PLYP: Schwabe/Grimme improved double hybrid
		"mB2PLYP":   ["HYB_GGA_XC_MB2PLYP"],
		# PWPB95: New Grimme double hybrid [94]
		"PWPB95":    ["HYB_MGGA_XC_MPW1B95"]
            }
        nomadNames = functionalMap.get(functional)
        if not nomadNames:
            raise Exception("Unhandled xc functional %s found" % functional)
        for name in nomadNames:
            s = backend.openSection("section_XC_functionals")
            backend.addValue('XC_functional_name', name)
            backend.closeSection("section_XC_functionals", s)

    def onClose_section_method(self, backend, gIndex, value):
        if value["electronic_structure_method"][-1] == "TDDFT":
            gi = backend.openSection("section_method_to_method_refs")
            backend.addValue("method_to_method_ref", (gIndex - 1))
            backend.addValue("method_to_method_kind", 'starting_point')
            backend.closeSection("section_method_to_method_refs", gi)

            gi = backend.openSection("section_calculation_to_calculation_refs")
            backend.addValue("calculation_to_calculation_ref", (self.CalculationGIndex - 1))
            backend.addValue("calculation_to_calculation_kind", 'starting_point')
            backend.closeSection("section_calculation_to_calculation_refs", gi)

    def onOpen_section_single_configuration_calculation(self, backend, gIndex, value):
        self.CalculationGIndex = gIndex

    def onClose_section_eigenvalues(self, backend, gIndex, value):
        occupations = np.array(value["x_orca_orbital_occupation_nb"])
        eigenvalues = np.array(value["x_orca_orbital_energy"])

        number_of_eigenvalues = len(eigenvalues)
        backend.addValue("number_of_eigenvalues", number_of_eigenvalues)
        backend.addValue("number_of_eigenvalues_kpoints", 1)

        occupations = np.array(value["x_orca_orbital_occupation_nb"])
        eigenvalues = np.array(value["x_orca_orbital_energy"])

        backend.addArrayValues("eigenvalues_occupation", occupations.reshape([1, 1, number_of_eigenvalues]))

        backend.addArrayValues("eigenvalues_values", eigenvalues.reshape([1, 1, number_of_eigenvalues]))

    def onClose_section_excited_states(self, backend, gIndex, value):
        number_of_excited_states = len(value["x_orca_excitation_energy"])
        backend.addValue("number_of_excited_states", number_of_excited_states)

        backend.addArrayValues("excitation_energies", np.array(value["x_orca_excitation_energy"]))

        backend.addArrayValues("oscillator_strengths", np.array(value["x_orca_oscillator_strength"]))

        x = value["x_orca_transition_dipole_moment_x"]
        y = value["x_orca_transition_dipole_moment_y"]
        z = value["x_orca_transition_dipole_moment_z"]
        tdm = np.zeros((number_of_excited_states,3), dtype=float)
        tdm[:, 0] = x
        tdm[:, 1] = y
        tdm[:, 2] = z
        backend.addArrayValues("transition_dipole_moments", tdm)

    def onClose_section_run(self, backend, gIndex, value):
        versionPieces = [value[x] for x in ["x_orca_program_version", "x_orca_program_svn", "x_orca_program_compilation_date", "x_orca_program_compilation_time"]]
        version = []
        for v in versionPieces:
            if v:
                version.append(v[0])
        if version:
            backend.addValue("program_version", " ".join(version))

##########################################################
############    [2] MAIN PARSER STARTS HERE   ############
##########################################################

def build_OrcaMainFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the main output file of ORCA.

    First, several subMatchers are defined, which are then used to piece together
    the final SimpleMatcher.
    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses main file of ORCA.
    """
#
# a) SimpleMatcher for header and ORCA version:
# *********************************************
#
    return SM(
        name = 'root',
        weak = True,
        startReStr = r"\s*\* O  R  C  A \*\s*",
        forwardMatch = True,
        sections = ["section_run"],
        fixedStartValues = {"program_name": "ORCA", "program_basis_set_type": "Gaussians" },
        subMatchers = [
            SM(name = 'ProgramHeader',
               startReStr = r"\s*\* O   ?R   ?C   ?A \*\s*",
               subMatchers = [
                   SM(r"\s*Program Version\s*(?P<x_orca_program_version>[0-9a-zA-Z_.].*)"),
                   SM(r"\s*\(SVN:\s*\$(?P<x_orca_program_svn>[^$]+)\$\)\s*"),
                   SM(r"\s*\(\$Date\:\s*(?P<x_orca_program_compilation_date>[0-9a-zA-Z].+?)\s*\$\)")
                   ]),
            buildSinglePointMatcher(),
            buildGeoOptMatcher(),
            buildMp2Matcher(),
            buildCIMatcher(),
            buildTDDFTMatcher()
            ])

def buildSinglePointMatcher():
#
# b) SimpleMatcher for Single Point Calculation:
# **********************************************
#
  return SM(name = 'SinglePointCalculation',
     startReStr = r"\s*\* Single Point Calculation \*\s",
     sections = ["section_single_configuration_calculation"],
     subMatchers = buildSinglePointSubMatchers()
  )

def buildSinglePointSubMatchers():
    return [
       # Get atomic positions:
       SM(name = 'Atomic Coordinates',
          startReStr = r"CARTESIAN COORDINATES \(ANGSTROEM\)\s*",
          sections = ["section_system", "x_orca_atom_positions"],
          subMatchers = [
          SM(r"\s+(?P<x_orca_atom_labels>[a-zA-Z]+)\s+(?P<x_orca_atom_positions_x__angstrom>[-+0-9.]+)\s+(?P<x_orca_atom_positions_y__angstrom>[-+0-9.]+)\s+(?P<x_orca_atom_positions_z__angstrom>[-+0-9.]+)", repeats = True)
          ]
       ),
       # Get basis set information:
       SM(name = 'Basis set information',
          startReStr = r"BASIS SET INFORMATION\s*",
          sections = ["section_basis_set"],
          subMatchers = [
          # Atom labels and basis set:
          SM(r"\s*Group\s+[0-9]+\s+Type\s+(?P<x_orca_basis_set_atom_labels>[a-zA-Z]+)\s+:\s+(?P<x_orca_basis_set>[0-9a-z]+)\s+contracted\s+to\s+(?P<x_orca_basis_set_contracted>[0-9a-z]+)\s+pattern\s+\{[0-9/]+\}", repeats = True),
          # Auxiliary basis set information:
          SM(name = 'Auxiliary basis set information',
             startReStr = r"AUXILIARY BASIS SET INFORMATION\s*",
             sections = ["section_basis_set"],
             subMatchers = [
             SM(r"\s*Group\s+[0-9]+\s+Type\s+(?P<x_orca_basis_set_atom_labels>[a-zA-Z]+)\s+:\s+(?P<x_orca_auxiliary_basis_set>[0-9a-z]+)\s+contracted\s+to\s+(?P<x_orca_auxiliary_basis_set_contracted>[0-9a-z]+)\s+pattern\s+\{[0-9/]+\}", repeats = True)
             ]
          )
          ]
       ),
       # Basis set statistics and startup info:
       SM(name = 'Basis set statistics and startup info',
          startReStr = r"\s*BASIS SET STATISTICS AND STARTUP INFO\s*",
          sections = ["section_basis_set"],
          subMatchers = [
          # Gaussian basis set:
          SM(name = 'Gaussian basis set',
             startReStr = r"\s*Gaussian basis set:\s*",
             subMatchers = [
             SM(r"\s+# of primitive gaussian shells\s+\.\.\.\s+(?P<x_orca_nb_of_primitive_gaussian_shells>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+# of primitive gaussian functions\s+\.\.\.\s+(?P<x_orca_nb_of_primitive_gaussian_functions>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+# of contracted shells\s+\.\.\.\s+(?P<x_orca_nb_of_contracted_shells>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+# of contracted basis functions\s+\.\.\.\s+(?P<x_orca_nb_of_contracted_basis_functions>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+Highest angular momentum\s+\.\.\.\s+(?P<x_orca_highest_angular_moment>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+Maximum contraction depth\s+\.\.\.\s+(?P<x_orca_maximum_contraction_depth>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+# of primitive gaussian shells\s+\.\.\.\s+(?P<x_orca_nb_primitive_gaussian_shells>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             ]
          ),
          # Gaussian auxiliary basis set:
          SM(name = 'Gaussian auxiliary basis set',
             startReStr = r"\s*Auxiliary gaussian basis set:\s*",
             subMatchers = [
             SM(r"\s+# of primitive gaussian shells\s+\.\.\.\s+(?P<x_orca_nb_of_primitive_gaussian_shells_aux>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+# of primitive gaussian functions\s+\.\.\.\s+(?P<x_orca_nb_of_primitive_gaussian_functions_aux>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+# of contracted shells\s+\.\.\.\s+(?P<x_orca_nb_of_contracted_shells_aux>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+# of contracted aux-basis functions\s+\.\.\.\s+(?P<x_orca_nb_of_contracted_basis_functions_aux>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+Highest angular momentum\s+\.\.\.\s+(?P<x_orca_highest_angular_moment_aux>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+Maximum contraction depth\s+\.\.\.\s+(?P<x_orca_maximum_contraction_depth_aux>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             SM(r"\s+# of primitive gaussian shells\s+\.\.\.\s+(?P<x_orca_nb_primitive_gaussian_shells_aux>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             ]
          )
          ]
       ),
       # SCF Settings:
       SM(name = 'Orca SCF settings',
          startReStr = r"\s*ORCA SCF\s*",
          sections = ["section_method", "x_orca_section_functionals"],
          fixedStartValues = {
              "electronic_structure_method": "DFT"
          },
          subMatchers = [
          # A - For HF methods:
          SM(r"\s+Ab initio Hamiltonian\s+Method\s+\.\.\.\s+(?P<x_orca_XC_functional_type>[-+0-9a-zA-Z()]+)"),
          # B - For DFT methods:
          SM(r"\s+Density Functional\s+Method\s+\.\.\.\s+(?P<x_orca_XC_functional_type>[a-zA-Z()]+)"),
          SM(r"\s+Exchange Functional\s+Exchange\s+\.\.\.\s+(?P<x_orca_exchange_functional>[a-zA-Z0-9]+)"),
          SM(r"\s+X-Alpha parameter\s+XAlpha\s+\.\.\.\s+(?P<x_orca_xalpha_param>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Becke's b parameter\s+XBeta\s+\.\.\.\s+(?P<x_orca_beckes_beta_param>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Correlation Functional\s+Correlation\s+\.\.\.\s+(?P<x_orca_correl_functional>[a-zA-Z0-9]+)"),
          SM(r"\s+LDA part of GGA corr\.\s+LDAOpt\s+\.\.\.\s+(?P<x_orca_lda_part_of_gga_corr>[a-zA-Z-+0-9]+)"),
          # Common settings:
          SM(r"\s+Scalar relativistic method\s+\.\.\.\s+(?P<x_orca_scalar_relativistic_method>[a-zA-Z-+0-9]+)"),
          SM(r"\s+Speed of light used\s+Velit\s+\.\.\.\s+(?P<x_orca_speed_of_light_used>[0-9.]+)"),
          SM(r"\s+Hartree-Fock type\s+HFTyps+\.\.\.\s+(?P<x_orca_hf_type>[a-zA-Z]+)"),
          SM(r"\s+Total Charge\s+Charge\s+\.\.\.\s+(?P<x_orca_total_charge>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Multiplicity\s+Mult\s+\.\.\.\s+(?P<x_orca_multiplicity>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Number of Electrons\s+NEL\s+\.\.\.\s+(?P<x_orca_nelectrons>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Nuclear Repulsion\s+ENuc\s+\.\.\.\s+(?P<x_orca_nuclear_repulsion__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          # Convergence Tolerance:
          SM(r"\s+Convergence Check Mode ConvCheckMode\s*\.\.\.\s+(?P<x_orca_convergence_check_mode>[a-zA-Z-+0-9.eEdD_-]+)"),
          SM(r"\s+Energy Change\s+TolE\s*\.\.\.\s+(?P<x_orca_energy_change_tolerance__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+1-El\. energy change\s*\.\.\.\s+(?P<x_orca_1_elect_energy_change__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          # DFT Grid generation:
          SM(r"\s+General Integration Accuracy\s+IntAcc\s*\.\.\.\s+(?P<x_orca_gral_integ_accuracy>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Radial Grid Type\s+RadialGrid\s*\.\.\.\s+(?P<x_orca_radial_grid_type>[a-zA-Z-_]+)"),
          SM(r"\s+Angular Grid \(max\. acc\.\)\s+AngularGrid\s*\.\.\.\s+(?P<x_orca_angular_grid>[a-zA-Z-+0-9.eEdD-_]+)"),
          SM(r"\s+Angular grid pruning method\s+GridPruning\s*\.\.\.\s+(?P<x_orca_grid_pruning_method>[a-zA-Z-+0-9.eEdD-_]+)"),
          SM(r"\s+Weight generation scheme\s*WeightScheme\s*\.\.\.\s+(?P<x_orca_weight_gener_scheme>[a-zA-Z0-9]+)"),
          SM(r"\s+Basis function cutoff\s+BFCut\s*\.\.\.\s+(?P<x_orca_basis_fn_cutoff>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Integration weight cutoff\s+WCut\s*\.\.\.\s+(?P<x_orca_integr_weight_cutoff>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+# of grid points \(after initial pruning\)\s+WCut\s*\.\.\.\s+(?P<x_orca_nb_grid_pts_after_initial_pruning>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+# of grid points \(after weights\+screening\)\s*\.\.\.\s+(?P<x_orca_nb_grid_pts_after_weights_screening>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Total number of grid points\s*\.\.\.\s+(?P<x_orca_total_nb_grid_pts>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Total number of batches\s*\.\.\.\s+(?P<x_orca_total_nb_batches>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Average number of points per batch\s*\.\.\.\s+(?P<x_orca_avg_nb_points_per_batch>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Average number of grid points per atom\s*\.\.\.\s+(?P<x_orca_avg_nb_grid_pts_per_atom>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)")
          ]
       ),
       # SCF iterations:
       SM(name = 'Orca SCF iterations',
          startReStr = r"\s*SCF ITERATIONS\s*",
          subMatchers = [
          SM(r"\s*(?P<x_orca_iteration_nb>[0-9]+)\s+(?P<energy_total_scf_iteration__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)",
             sections = ["section_scf_iteration"],
             repeats = True),
          ]
       ),
#      *****************************************************
#      *                     SUCCESS                       *
#      *           SCF CONVERGED AFTER  XY CYCLES          *
#      *****************************************************
       SM(name = 'Final step after convergence',
          startReStr = r"Setting up the final grid:",
          sections = ["section_scf_iteration", "section_basis_set"],
          subMatchers = [
          # Final DFT Grid generation:
          SM(r"\s+General Integration Accuracy\s+IntAcc\s*\.\.\.\s+(?P<x_orca_gral_integ_accuracy_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Radial Grid Type\s+RadialGrid\s*\.\.\.\s+(?P<x_orca_radial_grid_type_final>[a-zA-Z-_]+)"),
          SM(r"\s+Angular Grid \(max\. acc\.\)\s+AngularGrid\s*\.\.\.\s+(?P<x_orca_angular_grid_final>[a-zA-Z-+0-9.eEdD-_]+)"),
          SM(r"\s+Angular grid pruning method\s+GridPruning\s*\.\.\.\s+(?P<x_orca_grid_pruning_method_final>[a-zA-Z-+0-9.eEdD-_]+)"),
          SM(r"\s+Weight generation scheme\s*WeightScheme\s*\.\.\.\s+(?P<x_orca_weight_gener_scheme_final>[a-zA-Z0-9]+)"),
          SM(r"\s+Basis function cutoff\s+BFCut\s*\.\.\.\s+(?P<x_orca_basis_fn_cutoff_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Integration weight cutoff\s+WCut\s*\.\.\.\s+(?P<x_orca_integr_weight_cutoff_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+# of grid points \(after initial pruning\)\s+WCut\s*\.\.\.\s+(?P<x_orca_nb_grid_pts_after_initial_pruning_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+# of grid points \(after weights\+screening\)\s*\.\.\.\s+(?P<x_orca_nb_grid_pts_after_weights_screening_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Total number of grid points\s*\.\.\.\s+(?P<x_orca_total_nb_grid_pts_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Total number of batches\s*\.\.\.\s+(?P<x_orca_total_nb_batches_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Average number of points per batch\s*\.\.\.\s+(?P<x_orca_avg_nb_points_per_batch_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s+Average number of grid points per atom\s*\.\.\.\s+(?P<x_orca_avg_nb_grid_pts_per_atom_final>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          # Final SCF total Energy:
          SM(name = 'Total Energy',
             startReStr = r"\s*TOTAL SCF ENERGY\s*",
             sections = [],
             subMatchers = [
             SM(r"\s*Total Energy\s+:\s+(?P<energy_total__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             # Energy Components:
             SM(name = 'Energy Components',
                startReStr = r"\s*Components:\s*",
                subMatchers = [
                SM(r"\s*Nuclear Repulsion\s*:\s+(?P<x_orca_nuc_repulsion__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*Electronic Energy\s*:\s+(?P<x_orca_elec_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*One Electron Energy:\s+(?P<x_orca_one_elec_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*Two Electron Energy:\s+(?P<x_orca_two_elec_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                # Virial Components:
                SM(r"\s*Potential Energy\s*:\s+(?P<x_orca_potential_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*Kinetic Energy\s*:\s+(?P<x_orca_kinetc_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*Virial Ratio\s*:\s+(?P<x_orca_virial_ratio>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                # DFT Components:
                SM(r"\s*N\(Alpha\)\s*:\s+(?P<x_orca_nb_elect_alpha_channel>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*N\(Beta\)\s*:\s+(?P<x_orca_nb_elect_beta_channel>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*N\(Total\)\s*:\s+(?P<x_orca_nb_elect_total>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*E\(X\)\s*:\s+(?P<x_orca_exchange_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*E\(C\)\s*:\s+(?P<x_orca_correlation_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
                SM(r"\s*E\(XC\)\s*:\s+(?P<x_orca_exchange_correlation_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)")
                ]
             )
             ]
          ),
          # Final SCF convergence:
          SM(r"\s*Last Energy change\s+\.\.\.\s+(?P<x_orca_last_energy_change__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+Tolerance :\s*(?P<x_orca_last_energy_change_tolerance__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*Last MAX-Density change\s+\.\.\.\s+(?P<x_orca_last_max_density_change__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+Tolerance :\s*(?P<x_orca_last_max_density_tolerance__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*Last RMS-Density change\s+\.\.\.\s+(?P<x_orca_last_rms_density_change__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+Tolerance :\s*(?P<x_orca_last_rms_density_tolerance__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)")
          ]
       ),
       # Orbitals Energies and Mulliken population analysis:
       SM(name = 'Orbital Energies',
          startReStr = r"\s*ORBITAL ENERGIES\s*",
          sections = ["section_dos", "section_eigenvalues"],
          subMatchers = [
          SM(r"\s*(?P<x_orca_orbital_nb>[0-9]+)\s+(?P<x_orca_orbital_occupation_nb>[-+.0-9]+)\s+(?P<x_orca_orbital_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)", repeats = True),
          # Mulliken population analysis:
          SM(name = 'Mulliken population analysis',
             startReStr = r"\s*\* MULLIKEN POPULATION ANALYSIS \*\s*",
             sections = [],
             subMatchers = [
             SM(r"\s*(?P<x_orca_atom_nb>[0-9]+)\s+(?P<x_orca_atom_species>[a-zA-Z]+):\s+(?P<x_orca_mulliken_atom_charge>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)", repeats = True),
             SM(r"\s*Sum of atomic charges:\s*(?P<x_orca_mulliken_total_charge>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
             # Mulliken reduced orbital charges (mroc):
             SM(r"\s*(?P<x_orca_atom_nb_mroc>[0-9]+)\s+(?P<x_orca_atom_species_mroc>[a-zA-Z]+)(?P<x_orca_atom_orbital_mroc>[-+0-9a-zA-Z]+)\s*:\s*(?P<x_orca_mulliken_partial_orbital_charge_mroc>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)", repeats = True)
             ]
          )
          ]
       ),
       # Time table:
       SM(name = 'timings',
          startReStr = r"\s*TIMINGS\s*",
          subMatchers = [
          SM(r"\s*Total SCF time:\s+(?P<x_orca_total_days_time>[0-9]+) days (?P<x_orca_total_hours_time>[0-9]+) hours (?P<x_orca_total_mins_time>[0-9]+) min (?P<x_orca_total_secs_time>[0-9]+) sec"),
          SM(r"\s*Total time\s*\.\.\.\.\s*(?P<x_orca_final_time>[0-9.]+) sec"),
          SM(r"\s*Sum of individual times\s*\.\.\.\.\s*(?P<x_orca_sum_individual_times>[0-9.]+) sec"),
          SM(r"\s*Fock matrix formation\s*\.\.\.\.\s*(?P<x_orca_fock_matrix_formation>[0-9.]+) sec"),
          SM(r"\s*Coulomb formation\s*\.\.\.\.\s*(?P<x_orca_coulomb_formation>[0-9.]+) sec"),
          SM(r"\s*Split-RI-J\s*\.\.\.\.\s*(?P<x_orca_split_rj>[0-9.]+) sec"),
          SM(r"\s*XC integration\s*\.\.\.\.\s*(?P<x_orca_xc_integration>[0-9.]+) sec"),
          SM(r"\s*Basis function eval\.\s*\.\.\.\.\s*(?P<x_orca_basis_fn_evaluation>[0-9.]+) sec"),
          SM(r"\s*Density eval\.\s*\.\.\.\.\s*(?P<x_orca_density_evaluation>[0-9.]+) sec"),
          SM(r"\s*XC-Functional eval\.\s*\.\.\.\.\s*(?P<x_orca_xc_functional_evaluation>[0-9.]+) sec"),
          SM(r"\s*XC-Potential eval\.\s*\.\.\.\.\s*(?P<x_orca_potential_evaluation>[0-9.]+) sec"),
          SM(r"\s*Diagonalization\s*\.\.\.\.\s*(?P<x_orca_diagonalization>[0-9.]+) sec"),
          SM(r"\s*Density matrix formation\s*\.\.\.\.\s*(?P<x_orca_density_matrix_formation>[0-9.]+) sec"),
          SM(r"\s*Population analysis\s*\.\.\.\.\s*(?P<x_orca_population_analysis>[0-9.]+) sec"),
          SM(r"\s*Initial guess\s*\.\.\.\.\s*(?P<x_orca_initial_guess>[0-9.]+) sec"),
          SM(r"\s*Orbital Transformation\s*\.\.\.\.\s*(?P<x_orca_orbital_transformation>[0-9.]+) sec"),
          SM(r"\s*Orbital Orthonormalization\s*\.\.\.\.\s*(?P<x_orca_orbital_orthonormalization>[0-9.]+) sec"),
          SM(r"\s*DIIS solution\s*\.\.\.\.\s*(?P<x_orca_diis_solution>[0-9.]+) sec"),
          SM(r"\s*Grid generation\s*\.\.\.\.\s*(?P<x_orca_grid_generation>[0-9.]+) sec")
          ]
       )]

def buildGeoOptMatcher():
#
# c) SimpleMatcher for geometry optimization:
# *******************************************
#
       return SM(name = 'Geometry optimization',
          startReStr = r"\s*\* Geometry Optimization Run \*\s*",
          sections = ["section_sampling_method"],
          subMatchers = [
          # Geometry optimization settings:
          SM(r"Update method\s*(?P<x_orca_update_method>[a-zA-Z]+)\s+\.\.\.\.\s+(?P<x_orca_update_method_name>[a-zA-Z]+)"),
          SM(r"Choice of coordinates\s*(?P<x_orca_coords_choice>[a-zA-Z]+)\s+\.\.\.\.\s+(?P<x_orca_coords_choice_name>[a-zA-Z ]+)"),
          SM(r"Initial Hessian\s*(?P<x_orca_initial_hessian>[a-zA-Z]+)\s+\.\.\.\.\s+(?P<x_orca_initial_hessian_name>[a-zA-Z'\( \)]+)"),
          SM(r"Energy Change\s*(?P<x_orca_energy_change_tol>[a-zA-Z]+)\s+\.\.\.\.\s+(?P<x_orca_energy_change_tol_value__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) Eh"),
          SM(r"Max\. Gradient\s*(?P<x_orca_max_gradient_tol>[a-zA-Z]+)\s+\.\.\.\.\s+(?P<x_orca_max_gradient_tol_value__hartree_bohr_1>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) Eh/bohr"),
          SM(r"RMS Gradient\s*(?P<x_orca_rms_gradient_tol>[a-zA-Z]+)\s+\.\.\.\.\s+(?P<x_orca_rms_gradient_tol_value__hartree_bohr_1>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) Eh/bohr"),
          SM(r"Max\. Displacement\s*(?P<x_orca_max_displacement_tol>[a-zA-Z]+)\s+\.\.\.\.\s+(?P<x_orca_max_displacement_tol_value__bohr>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) bohr"),
          SM(r"RMS Displacement\s*(?P<x_orca_rms_displacement_tol>[a-zA-Z]+)\s+\.\.\.\.\s+(?P<x_orca_rms_displacement_tol_value__bohr>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) bohr"),
          SM(name="geoOptCycle",
             startReStr=r"\s*\*\s*GEOMETRY OPTIMIZATION CYCLE\s*(?P<x_orca_geo_opt_cycle>[0-9]+)\s*\*\s*",
             sections = ["section_single_configuration_calculation"],
             subMatchers = buildSinglePointSubMatchers()
          ),
          # Final geometry:
          # CARTESIAN COORDINATES (ANGSTROEM)
          SM(name = 'final geometry',
             startReStr = r"\s*\*\*\*\s*THE OPTIMIZATION HAS CONVERGED\s*\*\*\*\s*",
             sections = ["x_orca_final_geometry"],
             subMatchers = [
             SM(r"\s+(?P<x_orca_atom_labels_geo_opt>[a-zA-Z]+)\s+(?P<x_orca_atom_positions_x_geo_opt__angstrom>[-+0-9.]+)\s+(?P<x_orca_atom_positions_y_geo_opt__angstrom>[-+0-9.]+)\s+(?P<x_orca_atom_positions_z_geo_opt__angstrom>[-+0-9.]+)", repeats = True)
             ]
          ),
       # *** FINAL ENERGY EVALUATION AT THE STATIONARY POINT ***
       #
          SM(name = 'Final step after geometry convergence',
             startReStr = r"Setting up the final grid:",
             sections = ["section_single_configuration_calculation"],
             subMatchers = buildSinglePointSubMatchers()
         )
        ]
       )

def buildMp2Matcher():
#
# d) Post-processing calculations:
# ********************************
#
       # MP2 Calculation (post-proc):
    return SM(name = 'mp2',
          startReStr = r"\s*ORCA MP2 CALCULATION\s*",
              sections = ["section_method", "section_single_configuration_calculation"],
          fixedStartValues = {
              "electronic_structure_method": "MP2"
          },
          subMatchers = [
          SM(r"\s*Dimension of the basis\s*\.\.\.\s*(?P<x_orca_mp2_basis_dimension>[0-9.]+)"),
          SM(r"\s*Overall scaling of the MP2 energy\s*\.\.\.\s*(?P<x_orca_scaling_mp2_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) Eh"),
          SM(r"\s*Dimension of the aux-basis\s*\.\.\.\s*(?P<x_orca_mp2_aux_basis_dimension>[0-9.]+)"),
          SM(r"\s*RI-MP2 CORRELATION ENERGY:\s*(?P<energy_method_current__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) Eh"),
          SM(r"\s*MP2 TOTAL ENERGY:\s*(?P<energy_total__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) Eh")
          ]
       )
def buildCIMatcher():
       # Driven CI (post-proc):
    return SM(name = 'CI',
          startReStr = r"\s*ORCA-MATRIX DRIVEN CI\s*",
          sections = ["section_method", "section_single_configuration_calculation"],
#          fixedStartValues = {
#              "electronic_structure_method": "CI"
#          },
          subMatchers = [
          SM(r"\s*Correlation treatment\s*\.\.\.\s*(?P<electronic_structure_method>[()a-zA-Z0-9.]+)"),
          SM(r"\s*Single excitations\s*\.\.\.\s*(?P<x_orca_single_excitations_on_off>[a-zA-Z]+)"),
          SM(r"\s*Orbital optimization\s*\.\.\.\s*(?P<x_orca_orbital_opt_on_off>[a-zA-Z]+)"),
          SM(r"\s*Calculation of Z vector\s*\.\.\.\s*(?P<x_orca_z_vector_calc_on_off>[a-zA-Z]+)"),
          SM(r"\s*Calculation of Brueckner orbitals\s*\.\.\.\s*(?P<x_orca_Brueckner_orbitals_calc_on_off>[a-zA-Z]+)"),
          SM(r"\s*Perturbative triple excitations\s*\.\.\.\s*(?P<x_orca_perturbative_triple_excitations_on_off>[a-zA-Z]+)"),
          SM(r"\s*Calculation of F12 correction\s*\.\.\.\s*(?P<x_orca_f12_correction_on_off>[a-zA-Z]+)"),
          SM(r"\s*Frozen core treatment\s*\.\.\.\s*(?P<x_orca_frozen_core_treatment>[0-9.a-zA-Z( )]+)"),
          SM(r"\s*Reference Wavefunction\s*\.\.\.\s*(?P<x_orca_reference_wave_function>[0-9.a-zA-Z]+)"),
          SM(r"\s*Number of AO's\s*\.\.\.\s*(?P<x_orca_nb_of_atomic_orbitals>[0-9]+)"),
          SM(r"\s*Number of electrons\s*\.\.\.\s*(?P<x_orca_nb_of_electrons>[0-9]+)"),
          SM(r"\s*Number of correlated electrons\s*\.\.\.\s*(?P<x_orca_nb_of_correlated_electrons>[0-9]+)"),
          SM(r"\s*Integral transformation\s*\.\.\.\s*(?P<x_orca_integral_transformation>[a-zA-Z( )]+)"),
          SM(r"\s*K\(C\) Formation\s*\.\.\.\s*(?P<x_orca_K_C_formation>[+-a-zA-Z( )]+)"),
          SM(r"\s*Convergence tolerance \(max\. residuum\)\s*\.\.\.\s*(?P<x_orca_convergence_tol_max_residuum>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*Level shift for amplitude update\s*\.\.\.\s*(?P<x_orca_level_shift_amplitude_update>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          # Partial Coulomb Transformation:
          SM(r"\s*Transformation type\s*\.\.\.\s*(?P<x_orca_coulomb_transformation_type>[a-zA-Z ()]+)"),
          SM(r"\s*Dimension of the basis\s*\.\.\.\s*(?P<x_orca_coulomb_transformation_dimension_basis>[0-9]+)"),
          SM(r"\s*Number of internal alpha-MOs\s*\.\.\.\s*(?P<x_orca_nb_internal_alpha_mol_orbitals>[-+0-9( )]+)"),
          SM(r"\s*Number of internal beta-MOs\s*\.\.\.\s*(?P<x_orca_nb_internal_beta_mol_orbitals>[-+0-9( )]+)"),
          SM(r"\s*Pair cutoff\s*\.\.\.\s*(?P<x_orca_pair_cutoff__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?) Eh"),
          SM(r"\s*AO-integral source\s*\.\.\.\s*(?P<x_orca_atomic_orbital_integral_source>[a-zA-Z]+)"),
          SM(r"\s*Integral package used\s*\.\.\.\s*(?P<x_orca_integral_package_used>[a-zA-Z]+)"),
          SM(r"\s*Number of Alpha-MO pairs included\s*\.\.\.\s*(?P<x_orca_nb_alpha_pairs_included>[0-9]+)"),
          SM(r"\s*Number of Beta-MO pairs included\s*\.\.\.\s*(?P<x_orca_nb_beta_pairs_included>[0-9]+)"),
          # Spin-unrestricted guess:
          SM(r"\s*EMP2\(aa\)=\s*(?P<x_orca_mp2_energy_spin_aa__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*EMP2\(bb\)=\s*(?P<x_orca_mp2_energy_spin_bb__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*EMP2\(ab\)=\s*(?P<x_orca_mp2_energy_spin_ab__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*E\(0\)\s*\.\.\.\s*(?P<x_orca_mp2_initial_guess__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*E\(MP2\)\s*\.\.\.\s*(?P<x_orca_mp2_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*Initial E\(tot\)\s*\.\.\.\s*(?P<x_orca_mp2_total_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*<T\|T>\s*\.\.\.\s*(?P<x_orca_T_and_T_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*Number of pairs included\s*\.\.\.\s*(?P<x_orca_total_nb_pairs_included>[0-9]+)"),
          # iterations (e.g.:UHF COUPLED CLUSTER ITERATIONS):
          SM(r"\s*(?P<x_orca_ci_iteration_nb>[0-9]+)\s+(?P<x_orca_ci_total_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+(?P<x_orca_ci_correl_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+(?P<x_orca_ci_deltaE_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+(?P<x_orca_ci_residual_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+(?P<x_orca_ci_iteration_time>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+(?P<x_orca_ci_half_s_and_s_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)", repeats = True),
          # Final Coupled Cluster Energies:
          SM(r"\s*E\(CORR\)\s*\.\.\.\s*(?P<x_orca_ccsd_correlation_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*E\(TOT\)\s*\.\.\.\s*(?P<x_orca_ccsd_total_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*Singles norm <S\|S>\*\*1/2\s*\.\.\.\s*(?P<x_orca_single_norm_half_ss__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*T1 diagnostic\s*\.\.\.\s*(?P<x_orca_t1_diagnostic__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*Triples Correction (T)\s*\.\.\.\s*(?P<x_orca_ccsdt_total_triples_correction__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*alpha-alpha-alpha\s*\.\.\.\s*(?P<x_orca_ccsdt_aaa_triples_contribution__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*alpha-alpha-beta\s*\.\.\.\s*(?P<x_orca_ccsdt_aab_triples_contribution__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*alpha-beta -beta\s*\.\.\.\s*(?P<x_orca_ccsdt_aba_triples_contribution__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*beta -beta -beta\s*\.\.\.\s*(?P<x_orca_ccsdt_bbb_triples_contribution__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*Final correlation energy\s*\.\.\.\s*(?P<x_orca_ccsdt_final_corr_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*E\(CCSD\)\s*\.\.\.\s*(?P<x_orca_ccsd_final_energy__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"),
          SM(r"\s*E\(CCSD\(T\)\)\s*\.\.\.\s*(?P<energy_total__hartree>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)")
          ]
       )

def buildTDDFTMatcher():
    # TDDFT Calculation (post-proc):
    return SM(name='tddft',
              startReStr=r"\s*ORCA TD-DFT/TDA CALCULATION\s*",
              sections=["section_single_configuration_calculation", "section_method", "section_excited_states"],
              fixedStartValues={
                  "electronic_structure_method": "TDDFT"
              },
              subMatchers=[
              SM(r"\s*ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS\s*"),
              SM(r"\s*[0-9]+\s+(?P<x_orca_excitation_energy__inversecm>[-+.0-9]+)\s+[-+.0-9]+\s+(?P<x_orca_oscillator_strength>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)"
                 r"\s+[-+.0-9]+\s+(?P<x_orca_transition_dipole_moment_x__bohr>[-+.0-9]+)\s+(?P<x_orca_transition_dipole_moment_y__bohr>[-+.0-9]+)"
                 r"\s+(?P<x_orca_transition_dipole_moment_z__bohr>[-+.0-9]+)\s*", repeats = True),
              SM(r"\s*ABSORPTION SPECTRUM VIA TRANSITION VELOCITY DIPOLE MOMENTS\s*")
              ]
        )

       # Here new stuff:
#       SM(name = '',
#          startReStr = r"",
#          sections = [],
#          subMatchers = [
#          SM(),
#          SM(),
#          SM()
#          ]
#       ),
#
# c) SimpleMatcher for Calculation Including Atomic Optimizations:
# ****************************************************************
#

mainFileDescription = build_OrcaMainFileSimpleMatcher()



#
# This is the rest of Fawzi's code:
# *********************************
#
# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json

parserInfo = {
  "name": "sample_parser",
  "version": "1.0"
}

class OrcaParser():
   """ A proper class envolop for running this parser from within python. """
   def __init__(self, backend, **kwargs):
       self.backend_factory = backend

   def parse(self, mainfile):
       from unittest.mock import patch
       logging.info('orca parser started')
       logging.getLogger('nomadcore').setLevel(logging.WARNING)
       backend = self.backend_factory("orca.nomadmetainfo.json")
       with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
           mainFunction(
               mainFileDescription,
               None,
               parserInfo,
               superContext=OrcaContext(),
               superBackend=backend)

       return backend



from typing import Union, Dict, List

from optimade.models import Species as OptimadeStructureSpecies
from optimade.models import StructureResource as OptimadeStructure

from optimade.adapters.structures.utils import pad_positions

try:
    from pymatgen import Structure, Molecule

except (ImportError, ModuleNotFoundError):
    from warnings import warn

    Structure = None
    Molecule = None
    PYMATGEN_NOT_FOUND = "Pymatgen not found, cannot convert structure to a pymatgen Structure or Molecule"


__all__ = ("get_pymatgen",)


def get_pymatgen(optimade_structure: OptimadeStructure) -> Union[Structure, Molecule]:
    """ Get pymatgen Structure or Molecule from OPTIMADE structure

    :param optimade_structure: OPTIMADE structure
    :return: pymatgen.Structure , pymatgen.Molecule
    """
    if globals().get("Structure", None) is None:
        warn(PYMATGEN_NOT_FOUND)
        return None

    if all(optimade_structure.attributes.dimension_types):
        return _get_structure(optimade_structure)

    return _get_molecule(optimade_structure)


def _get_structure(optimade_structure: OptimadeStructure) -> Structure:
    """Create pymatgen Structure from OPTIMADE structure"""

    attributes = optimade_structure.attributes

    cartesian_site_positions, _ = pad_positions(attributes.cartesian_site_positions)

    return Structure(
        lattice=attributes.lattice_vectors,
        species=_pymatgen_species(
            nsites=attributes.nsites,
            species=attributes.species,
            species_at_sites=attributes.species_at_sites,
        ),
        coords=cartesian_site_positions,
        coords_are_cartesian=True,
    )


def _get_molecule(optimade_structure: OptimadeStructure) -> Molecule:
    """Create pymatgen Molecule from OPTIMADE structure"""

    attributes = optimade_structure.attributes

    cartesian_site_positions, _ = pad_positions(attributes.cartesian_site_positions)

    return Molecule(
        species=_pymatgen_species(
            nsites=attributes.nsites,
            species=attributes.species,
            species_at_sites=attributes.species_at_sites,
        ),
        coords=cartesian_site_positions,
    )


def _pymatgen_species(
    nsites: int, species: List[OptimadeStructureSpecies], species_at_sites: List[str]
) -> List[Dict[str, float]]:
    """
    Create list of {"symbol": "concentration"} per site for values to pymatgen species parameters.
    Remove vacancies, if they are present.
    """

    optimade_species = {_.name: _ for _ in species}

    pymatgen_species = []
    for site_number in range(nsites):
        species_name = species_at_sites[site_number]
        current_species = optimade_species[species_name]

        chemical_symbols = []
        concentration = []
        for index, symbol in enumerate(current_species.chemical_symbols):
            if symbol == "vacancy":
                # Skip. This is how pymatgen handles vacancies;
                # to not include them, while keeping the concentration in a site less than 1.
                continue
            else:
                chemical_symbols.append(symbol)
                concentration.append(current_species.concentration[index])

        pymatgen_species.append(dict(zip(chemical_symbols, concentration)))

    return pymatgen_species



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
The NOMAD meta-info allows to define schemas for physics data independent of the used
storage format. It allows to define physics quantities with types, complex shapes
(vetors, matrices, etc.), units, links, and descriptions. It allows to organize large
amounts of these quantities in containment hierarchies of extendable sections, references
between sections, and additional quantity categories.

NOMAD uses the meta-info to define all archive data, repository meta-data, (and encyclopedia
data). The meta-info provides a convenient Python interface to create,
manipulate, and access data. We also use it to map data to various storage formats,
including JSON, (HDF5), mongodb, and elastic search.

Starting example
----------------

.. code-block:: python

    from nomad.metainfo import MSection, Quantity, SubSection, Units

    class System(MSection):
        \'\'\'
        A system section includes all quantities that describe a single a simulated
        system (a.k.a. geometry).
        \'\'\'

        n_atoms = Quantity(
            type=int, description=\'\'\'
            A Defines the number of atoms in the system.
            \'\'\')

        atom_labels = Quantity(type=MEnum(ase.data.chemical_symbols), shape['n_atoms'])
        atom_positions = Quantity(type=float, shape=['n_atoms', 3], unit=Units.m)
        simulation_cell = Quantity(type=float, shape=[3, 3], unit=Units.m)
        pbc = Quantity(type=bool, shape=[3])

    class Run(MSection):
        systems = SubSection(sub_section=System, repeats=True)


We define simple metainfo schema with two `sections` called ``System`` and ``Run``. Sections
allow to organize related data into,  well, `sections`. Each section can have two types of
properties: `quantities` and `sub-sections`. Sections and their properties are defined with
Python classes and their attributes.

Each `quantity` defines a piece of data. Basic quantity attributes are its `type`, `shape`,
`unit`, and `description`.

`Sub-sections` allow to place section into each other and there allow to form containment
hierarchies or sections and the respective data in them. Basic sub-section attributes are
`sub_section`(i.e. a reference to the section definition of the sub-section) and `repeats`
(determines if a sub-section can be contained once or multiple times).

The above simply defines a schema, to use the schema and create actual data, we have to
instantiate the above classes:

.. code-block:: python

    run = Run()
    system = run.m_create(System)
    system.n_atoms = 3
    system.atom_labels = ['H', 'H', 'O']

    print(system.atom_labels)
    print(n_atoms = 3)


Section `instances` can be used like regular Python objects: quantities and sub-sections
can be set and access like any other Python attribute. Special meta-info methods, starting
with ``m_`` allow us to realize more complex semantics. For example ``m_create`` will
instantiate a sub-section and add it to the `parent` section in one step.

Another example for an ``m_``-method is:

.. code-block:: python

    run.m_to_json(indent=2)

This will serialize the data into JSON:

.. code-block:: JSON

    {
        "m_def" = "Run",
        "systems": [
            {
                "n_atoms" = 3,
                "atom_labels" = [
                    "H",
                    "H",
                    "O"
                ]
            }
        ]
    }

Definitions
-----------

.. autoclass:: Definition


Quantities
----------

.. autoclass:: Quantity

.. _metainfo-sections:

Sections
--------

With sections it is paramount to always be clear what is talked about. The lose
term `section` can reference one of the following three:

* `section definition`
    Which is a Python object that represents the definition of
    a section, its sub-sections and quantities. `Section definitions` should not be not
    written directly. `Section definitions` are objects of :class:`Section`.
* `secton class`
    Which is a Python class and :class:`MSection` decendant that is
    used to express a `section defintion` in Python. Each `section class` is tightly
    associated with its `section definition`. The `section definition` can be access
    with the class attribute ``m_def``. The `section definition` is automatically created
    from the `section class` upon defining the class through metaclass vodoo.
* `section instance`
    The instance (object) of a `section class`, it `follows` the
    definition associated with the instantiated `section class`. The followed
    section definition can be accessed with the object attribute ``m_def``.

A `section class` looks like this:

.. code-block:: python

    class SectionName(BaseSection):
        \'\'\' Section description \'\'\'
        m_def = Section(**section_attributes)

        quantity_name = Quantity(**quantity_attributes)
        sub_section_name = SubSection(**sub_section_attributes)

The various Python elements of this class are mapped to a respective *section definition*.
The ``SectionName`` becomes the *name*. The
``BaseSection`` is either :class:`MSection` or another *section class*. The
``section_attributes`` become additional attributes of the `section definition`. The
various ``Quantity`` and ``SubSection`` become the *quantities* and *sub sections*.

Each *section class* has to directly or indirectly extend :class:`MSection`. This will
provided certain class and object features to all *section classes* and all *section instances*.
Read :ref:metainfo-reflection to learn more.

.. autoclass:: Section

Sub-Sections
------------

.. autoclass:: SubSection

.. _metainfo-categories:

Categories
----------

.. autoclass:: Quantity

In the old meta-info this was known as `abstract types`.

Categories are defined with Python classes that have :class:`MCategory` as base class.
Their name and description is taken from the class's name and docstring. An example
category looks like this:

.. code-block:: python

    class CategoryName(MCategory):
        \'\'\' Category description \'\'\'
        m_def = Category(links=['http://further.explanation.eu'], categories=[ParentCategory])

Packages
--------

.. autoclass:: Package

.. _metainfo-custom-types:


Environments
------------

.. autoclass:: Environment


Custom data types
-----------------

.. autoclass:: DataType
    :members:

.. autoclass:: MEnum

.. _metainfo-reflection:

Reflection and custom data storage
----------------------------------

When manipulating metainfo data in Python, all data is represented as Python objects, where
objects correspond to `section instance` and their attributes to `quantity values` or
`section instances` of sub-sections. By defining sections with `section classes` each
of these Python objects already has an interface that allows to get/set quantities and
sub-sections. But often this interface is too limited, or the specific section and
quantity definitions are unknown when writing code.

.. autoclass:: MSection
    :members:

.. autoclass:: MetainfoError
.. autoclass:: DeriveError
.. autoclass:: MetainfoReferenceError

.. _metainfo-urls:

References and metainfo URLs
----------------------------

When in Python memory, quantity values that reference other sections simply contain a
Python reference to the respective `section instance`. However, upon serializing/storing
metainfo data, these references have to be represented differently.

Currently this metainfo implementation only supports references within a single
section hierarchy (e.g. the same JSON file). References are stored as paths from the
root section, over sub-sections, to the references section. Each path segment is
the name of the sub-section or an index in a repeatable sub-section:
``/system/0/symmetry``.

References are automatically serialized by :py:meth:`MSection.m_to_dict`. When de-serializing
data with :py:meth:`MSection.m_from_dict` these references are not resolved right away,
because the references section might not yet be available. Instead references are stored
as :class:`MProxy` instances. These objects are automatically replaced by the referenced
object when a respective quantity is accessed.

.. autoclass:: MProxy

Resources
_________

.. autoclass:: MResource

A more complex example
----------------------

.. literalinclude:: ../nomad/metainfo/example.py
    :language: python

'''


from .metainfo import (
    MSectionBound,
    MSection,
    MCategory,
    Definition,
    Property,
    Quantity,
    SubSection,
    Section,
    Category,
    Package,
    Environment,
    MEnum,
    Datetime,
    Capitalized,
    MProxy,
    MetainfoError,
    DeriveError,
    MetainfoReferenceError,
    DataType,
    Reference,
    Datetime,
    JSON,
    MResource,
    m_package,
    Annotation,
    DefinitionAnnotation,
    SectionAnnotation,
    SectionProxy,
    derived,
    constraint,
    units)



"""
Module for unit conversion routines. Currently uses the Pint unit conversion
library (https://pint.readthedocs.org) to do the conversions.

Any new units and constants can be added to the text files "units.txt" and
"constants.txt".
"""
from __future__ import print_function
from builtins import str
from builtins import object
import re
import logging
from pint import UnitRegistry
import cachetools
import numpy as np
logger = logging.getLogger(__name__)
# disable warnings from pint
logging.getLogger("pint").setLevel(logging.ERROR)

# ureg = UnitRegistry(os.path.join(os.path.dirname(__file__), "units.txt"))
ureg = UnitRegistry()
ureg.define('forceAu = hartree / bohr')
ureg.define('inversecm = 1.9864475e-23 * joule')
ureg.define('mubohr = 9.274009994e-24 J / T')


@cachetools.cached(cache={})
def ureg_cached(unit):
    return ureg(unit)


def register_userdefined_quantity(quantity, units, value=1):
    """Registers a user defined quantity, valid until redefined.
    The value should be equal to value using units, with value defaulting to 1
    """
    if not re.match(r"usr[A-Za-z]*", quantity):
        raise Exception("User defined quantities should start with usr, and contain only letters")
    if "=" in units:
        raise Exception("Units should not contain an = sign")
    float(value)  # value must be a float number
    ureg.define(quantity + ' = ' + str(value) + " * " + units)


def convert_unit(value, unit, target_unit=None):
    """Converts the given value from the given units to the target units. For
    examples see the bottom section.

    Args:
        value: The numeric value to be converted. Accepts integers, floats,
            lists and numpy arrays
        unit: The units that the value is currently given in as a string. All
            units that have a corresponding declaration in the "units.txt" file
            and combinations like "meter*second**-2" are supported.
        target_unit: The target unit as string. Same rules as for the unit
            argument. If this argument is not given, SI units are assumed.

    Returns:
        The given value in the target units. returned as the same data type as
        the original values.
    """
    factor = get_multiplicative_factor(unit, target_unit)
    if factor is not None:
        try:
            if isinstance(value, (list, tuple)):
                return (np.array(value) * factor).tolist()
            else:
                return value * factor
        except Exception as e:
            logger.error('could not perform optimized unit conversion', exc_info=e)

    # Check that the unit is valid
    unit_def = ureg_cached(unit)
    if not unit_def:
        logger.error("Undefined unit given. Cannot do the conversion")
        return

    # If no target is specified, assume SI automatically
    if not target_unit:
        Q_ = ureg.Quantity
        pint_value = Q_(value, unit_def)
        converted_value = pint_value.to_base_units()  # Base units are defined in the "units.txt" file, and they are the SI units.
        return converted_value.magnitude

    else:
        # Check that the given target unit is valid
        target_unit_def = ureg_cached(target_unit)
        if not target_unit_def:
            logger.error("Undefined target unit given. Cannot do the conversion")
            return

        Q_ = ureg.Quantity
        pint_value = Q_(value, unit_def)
        converted_value = pint_value.to(target_unit_def)
        return converted_value.magnitude


@cachetools.cached(cache={})
def get_multiplicative_factor(unit, target_unit=None):
    '''
    An optimization for unit conversion. In 99% of the cases this can produce a scaling
    factor to convert from unit to target_unit (or base units) and the
    results can be cached.
    '''
    try:
        unit_factor, _ = ureg.get_base_units(ureg_cached(unit), check_nonmult=True)
        if target_unit is not None:
            target_factor, _ = ureg.get_base_units(ureg_cached(target_unit), check_nonmult=True)
        else:
            target_factor = 1

        return unit_factor / target_factor

    except Exception:
        return None


def convert_unit_function_immediate(unit, target_unit=None):
    """Returns a function that converts scalar floats from unit to target_unit
    All units need to be already known.

    For more details see the convert_unit function.
    Could be optimized a bit caching the pint quantities

    Args:
        unit: The units that the value is currently given in as a string. All
            units that have a corresponding declaration in the "units.txt" file
            and combinations like "meter*second**-2" are supported.
        target_unit: The target unit as string. Same rules as for the unit
            argument. If this argument is not given, SI units are assumed.

    Returns:
        The given value in the target units. returned as the same data type as
        the original values.
    """
    # Check that the dimensionality of the source and target units match.
    if target_unit is not None:
        source = ureg_cached(target_unit)
        source_dim = source.dimensionality
        target = ureg_cached(unit)
        target_dim = target.dimensionality
        if source_dim != target_dim:
            raise Exception("The dimensionality of unit '{}' does not match the dimensionality of unit '{}'. Cannot do the unit conversion.".format(unit, target_unit))

    return lambda x: convert_unit(x, unit, target_unit)


class LazyF(object):
    """helper class for lazy evaluation of conversion function"""
    def __init__(self, unit, target_unit):
        self.unit = unit
        self.target_unit = target_unit
        self.f = None

    def __call__(self, x):
        if self.f is not None:
            return self.f(x)
        else:
            self.f = convert_unit_function_immediate(self.unit, self.target_unit)
            return self.f(x)


def convert_unit_function(unit, target_unit=None):
    """Returns a function that converts scalar floats from unit to target_unit
    if any of the unit are user defined (usr*), then the conversion is done lazily
    at the first call (i.e. user defined conversions might be undefined when
    calling this)

    For more details see the convert_unit function.
    Could be optimized a bit caching the pint quantities

    Args:
        unit: The units that the value is currently given in as a string. All
            units that have a corresponding declaration in the "units.txt" file
            and combinations like "meter*second**-2" are supported.
        target_unit: The target unit as string. Same rules as for the unit
            argument. If this argument is not given, SI units are assumed.

    Returns:
        The given value in the target units. returned as the same data type as
        the original values.
    """
    if "usr" in unit:
        return LazyF(unit, target_unit)
    else:
        return convert_unit_function_immediate(unit, target_unit)


# Testing
if __name__ == "__main__":

    import numpy as np

    # Float
    a = 20
    unit = "angstrom"
    target = "m"
    a_c = convert_unit(a, unit, target)
    print(a_c)

    # Numpy arrays
    b = np.ones((3, 3, 3))
    unit = "angstrom"
    target = "meter"
    b_c = convert_unit(b, unit, target)
    print(b_c)

    # Operators
    c = 20
    unit = "eV*angstrom**-1"
    c_c = convert_unit(c, unit)
    print(c_c)

    # Lists
    d = [20, 10, 0]
    unit = "angstrom"
    target = "m"
    d_c = convert_unit(d, unit, target)
    print(d_c)

    # Temperature
    e = 25
    unit = "celsius"
    target = "kelvin"
    e_c = convert_unit(e, unit, target)
    print(e_c)



# pylint: disable=protected-access,pointless-statement,relative-beyond-top-level
import json
import os
import unittest

from pathlib import Path

from optimade.server.config import ServerConfig, CONFIG

from .utils import SetClient


class ConfigTests(unittest.TestCase):
    def test_env_variable(self):
        """Set OPTIMADE_DEBUG environment variable and check CONFIG picks up on it correctly"""
        os.environ["OPTIMADE_DEBUG"] = "true"
        CONFIG = ServerConfig()
        self.assertTrue(CONFIG.debug)

        os.environ.pop("OPTIMADE_DEBUG", None)
        CONFIG = ServerConfig()
        self.assertFalse(CONFIG.debug)

    def test_default_config_path(self):
        """Make sure the default config path works
        Expected default config path: PATH/TO/USER/HOMEDIR/.optimade.json
        """
        # Reset OPTIMADE_CONFIG_FILE
        original_OPTIMADE_CONFIG_FILE = os.environ.get("OPTIMADE_CONFIG_FILE", "")
        os.environ.pop("OPTIMADE_CONFIG_FILE")

        with open(
            Path(__file__).parent.parent.joinpath("test_config.json"), "r"
        ) as config_file:
            config = json.load(config_file)

        different_base_url = "http://something_you_will_never_think_of.com"
        config["base_url"] = different_base_url

        # Try-finally to make sure we don't overwrite possible existing `.optimade.json`
        original = Path.home().joinpath(".optimade.json")
        restore = False
        if original.exists():
            restore = True
            with open(original, "rb") as original_file:
                original_file_content = original_file.read()
        try:
            with open(
                Path.home().joinpath(".optimade.json"), "w"
            ) as default_config_file:
                json.dump(config, default_config_file)
            CONFIG = ServerConfig()
            self.assertEqual(
                CONFIG.base_url,
                different_base_url,
                f"\nDumped file content:\n{config}.\n\nLoaded CONFIG:\n{CONFIG}",
            )
        finally:
            if restore:
                with open(original, "wb") as original_file:
                    original_file.write(original_file_content)

        # Restore OPTIMADE_CONFIG_FILE
        os.environ["OPTIMADE_CONFIG_FILE"] = original_OPTIMADE_CONFIG_FILE


class TestRegularServerConfig(SetClient, unittest.TestCase):

    server = "regular"

    def test_debug_is_respected_when_off(self):
        """Make sure traceback is toggleable according to debug mode - here OFF

        TODO: This should be moved to a separate test file that tests the exception handlers.
        """
        if CONFIG.debug:
            CONFIG.debug = False

        response = self.client.get("/non/existent/path")
        self.assertEqual(
            response.status_code,
            404,
            msg=f"Request should have failed, but didn't: {response.json()}",
        )

        response = response.json()
        self.assertNotIn("data", response)
        self.assertIn("meta", response)

        self.assertNotIn(f"_{CONFIG.provider.prefix}_traceback", response["meta"])

    def test_debug_is_respected_when_on(self):
        """Make sure traceback is toggleable according to debug mode - here ON

        TODO: This should be moved to a separate test file that tests the exception handlers.
        """
        CONFIG.debug = True

        response = self.client.get("/non/existent/path")
        self.assertEqual(
            response.status_code,
            404,
            msg=f"Request should have failed, but didn't: {response.json()}",
        )

        response = response.json()
        self.assertNotIn("data", response)
        self.assertIn("meta", response)

        self.assertIn(f"_{CONFIG.provider.prefix}_traceback", response["meta"])


class TestRegularIndexServerConfig(SetClient, unittest.TestCase):

    server = "index"

    def test_debug_is_respected_when_off(self):
        """Make sure traceback is toggleable according to debug mode - here OFF

        TODO: This should be moved to a separate test file that tests the exception handlers.
        """
        if CONFIG.debug:
            CONFIG.debug = False

        response = self.client.get("/non/existent/path")
        self.assertEqual(
            response.status_code,
            404,
            msg=f"Request should have failed, but didn't: {response.json()}",
        )

        response = response.json()
        self.assertNotIn("data", response)
        self.assertIn("meta", response)

        self.assertNotIn(f"_{CONFIG.provider.prefix}_traceback", response["meta"])

    def test_debug_is_respected_when_on(self):
        """Make sure traceback is toggleable according to debug mode - here OFF

        TODO: This should be moved to a separate test file that tests the exception handlers.
        """
        CONFIG.debug = True

        response = self.client.get("/non/existent/path")
        self.assertEqual(
            response.status_code,
            404,
            msg=f"Request should have failed, but didn't: {response.json()}",
        )

        response = response.json()
        self.assertNotIn("data", response)
        self.assertIn("meta", response)

        self.assertIn(f"_{CONFIG.provider.prefix}_traceback", response["meta"])



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from LAMMPS.BaseClasses import AtomStyle
import warnings

    #
    # if style == 'body':
    #     return body(*args)
    #
    # elif style == 'template':
    #     return template(*args)
    #
    # elif style == 'hybrid':
    #
    #     list = hybrid()
    #     args_iter = args.__iter__()
    #
    #     while True:
    #         try:
    #             style = args_iter.__next__()
    #
    #             if style == 'body':
    #                 bstyle = args_iter.__next__()
    #                 nmin = args_iter.__next__()
    #                 nmax = args_iter.__next__()
    #                 list.append(body(bstyle, nmin, nmax))
    #
    #             elif style == 'template':
    #                 templat_id = args_iter.__next__()
    #                 list.append(template(templat_id))
    #
    #             else:
    #                 list.append(eval(style)())
    #
    #         except StopIteration:
    #             break
    #     return list
    # else:
    #     return eval(style)()



class Base(AtomStyle):
    pass


class Angle(AtomStyle):
    name = 'angle'
    pass

class Atomic(AtomStyle):
    name = 'atomic'

class Angle(AtomStyle):
    name = 'angle'
    pass
class Atomic(AtomStyle):
    name = 'atomic'
    pass

class Body(AtomStyle):
    name = 'body'

    def __init__(self, bstyle, Nmin, Nmax):
        self.bstyle = bstyle
        self.nmin = Nmin
        self.nmax = Nmax

class Bond(AtomStyle):
    name = 'bond'
    pass
class Charge(AtomStyle):
    name = 'charge'
    pass
class Dipole(AtomStyle):
    name = 'dipole'
    pass
class Dpd(AtomStyle):
    name = 'dpd'
    pass
class Electron(AtomStyle):
    name = 'electron'
    pass
class Ellipsoid(AtomStyle):
    name = 'ellipsoid'
    pass
class Full(AtomStyle):
    name = 'full'
    pass
class Line(AtomStyle):
    name = 'line'
    pass
class Meso(AtomStyle):
    name = 'meso'
    pass
class Molecular(AtomStyle):
    name = 'molecular'
    pass
class Peri(AtomStyle):
    name = 'peri'
    pass
class Smd(AtomStyle):
    name = 'smd'
    pass
class Sphere(AtomStyle):
    name = 'sphere'
    pass
class Tri(AtomStyle):
    name = 'tri'
    pass

class Wavepacket(AtomStyle):
    pass

class Template(AtomStyle):
    name = 'template'

    def __init__(self, id):
        self.id = id


class Hybrid(AtomStyle):
    name = 'hybrid'

    def __init__(self, *args):
        pass
    #     super().__init__()
    #     self.list = []
    #
    #     # create an itarator
    #     args_iter = args.__iter__()
    #     while True:
    #         try:
    #             style = args_iter.__next__()
    #
    #             if style == 'body':
    #                 bstyle = args_iter.__next__()
    #                 nmin = args_iter.__next__()
    #                 nmax = args_iter.__next__()
    #                 self.list.append(body(bstyle, nmin, nmax))
    #             elif style == 'template':
    #                 templat_id = args_iter.__next__()
    #                 self.list.append(template(templat_id))
    #             else:
    #                 self.list.append(eval(style))
    #
    #         except StopIteration:
    #             break

_STYLE_DICT = {
    'angle': Angle,
    'atomic': Atomic,
    'body': Body,
    'bond': Bond,
    'charge': Charge,
    'dipole': Dipole,
    'dpd': Dpd,
    'electron': Electron,
    'ellipsoid': Ellipsoid,
    'full': Full,
    'line': Line,
    'meso': Meso,
    'molecular': Molecular,
    'peri': Peri,
    'smd': Smd,
    'sphere': Sphere,
    'tri': Tri,
    'wavepacket': Wavepacket,
    'template': Template,
    'hybrid': Hybrid,
}


def atom_style(style, *args):

    cls = _STYLE_DICT.get(style, None)

    if not cls:
        print("Unsupported atom_style: {}".format(style))
        return None

    return cls(*args)




if __name__ == '__main__':
    a = Bond()
    print(a)

    a = atom_style('atomic')
    print(a)

    a = atom_style('bond')
    print(a)

    atom_style('full')
    atom_style('body', 'nparticle', 2, 10)
    atom_style('hybrid', 'charge', 'bond')
    atom_style('hybrid', 'charge', 'body', 'nparticle', 2, 5)
    atom_style('template', 'myMols')


# Copyright 2018-2018 Berk Onat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import os

PARSERNAME = "TINKER"
PROGRAMNAME = "tinker"
PARSERVERSION = "0.0.3"
PARSERMETANAME = PARSERNAME.lower()
PARSERTAG = 'x_' + PARSERMETANAME

PARSER_INFO_DEFAULT = {
        'name'   : PARSERMETANAME+'-parser',
        'version': PARSERVERSION
}

META_INFO_PATH = PARSERMETANAME+".nomadmetainfo.json"

LOGGER = logging.getLogger("nomad."+PROGRAMNAME+"Parser")

def set_excludeList(self):
    """Sets the exclude list for x_

    Returns:
        the list of names
    """
    excludelist = [
        PARSERTAG+'_mdin_verbatim_writeout',
        PARSERTAG+'_dumm_text',
        PARSERTAG+'_dummy',
        PARSERTAG+'_mdin_wt'
        ]
    excludelist.extend([PARSERTAG+'_mdin_file_%s' % fileNL.lower() for fileNL in self.fileDict.keys()])
    excludelist.extend([PARSERTAG+'_mdin_%s' % cntrlNL.lower() for cntrlNL in self.cntrlDict.keys()])
    excludelist.extend([PARSERTAG+'_mdin_%s' % ewaldNL.lower() for ewaldNL in self.ewaldDict.keys()])
    excludelist.extend([PARSERTAG+'_mdin_%s' % qmmmNL.lower() for qmmmNL in self.qmmmDict.keys()])
    return excludelist

def set_includeList():
    """Sets the include list for x_

    Returns:
        the list of names
    """
    includelist = []
    return includelist





# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import asyncio
from concurrent import futures as concurrent_futures

from nomad import config
from .admin import admin


@admin.group(help='Run a nomad service locally (outside docker).')
def run():
    pass


@run.command(help='Run the nomad development worker.')
def worker():
    run_worker()


@run.command(help='Run the nomad development app with all apis.')
@click.option('--debug', help='Does run flask in debug.', is_flag=True)
@click.option('--with-chaos', type=int, default=0, help='Enable a certain percentage of chaos.')
def app(debug: bool, with_chaos: int):
    config.services.api_chaos = with_chaos
    run_app(debug=debug)


def run_app(**kwargs):
    config.meta.service = 'app'
    from nomad import infrastructure
    from nomad.app.__main__ import run_dev_server
    infrastructure.setup()
    run_dev_server(port=config.services.api_port, **kwargs)


def run_worker():
    config.meta.service = 'worker'
    from nomad import processing
    processing.app.worker_main(['worker', '--loglevel=INFO', '-Q', 'celery,uploads,calcs'])


@run.command(help='Run both app and worker.')
def appworker():
    executor = concurrent_futures.ProcessPoolExecutor(2)
    loop = asyncio.get_event_loop()
    loop.run_in_executor(executor, run_app)
    loop.run_in_executor(executor, run_worker)



# Copyright 2016-2018 Ask Hjorth Larsen, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# All the space groups that GULP will print
sg = ['P 1             ','P -1            ','P 2             ','P 21            ',
      'C 2             ','P M             ','P C             ','C M             ','C C             ',
      'P 2/M           ','P 21/M          ','C 2/M           ','P 2/C           ','P 21/C          ',
      'C 2/C           ','P 2 2 2         ','P 2 2 21        ','P 21 21 2       ','P 21 21 21      ',
      'C 2 2 21        ','C 2 2 2         ','F 2 2 2         ','I 2 2 2         ','I 21 21 21      ',
      'P M M 2         ','P M C 21        ','P C C 2         ','P M A 2         ','P C A 21        ',
      'P N C 2         ','P M N 21        ','P B A 2         ','P N A 21        ','P N N 2         ',
      'C M M 2         ','C M C 21        ','C C C 2         ','A M M 2         ','A B M 2         ',
      'A M A 2         ','A B A 2         ','F M M 2         ','F D D 2         ','I M M 2         ',
      'I B A 2         ','I M A 2         ','P M M M         ','P N N N         ','P C C M         ',
      'P B A N         ','P M M A         ','P N N A         ','P M N A         ','P C C A         ',
      'P B A M         ','P C C N         ','P B C M         ','P N N M         ','P M M N         ',
      'P B C N         ','P B C A         ','P N M A         ','C M C M         ',
      'C M C A         ','C M M M         ','C C C M         ','C M M A         ','C C C A         ',
      'F M M M         ','F D D D         ','I M M M         ','I B A M         ','I B C A         ',
      'I M M A         ','P 4             ','P 41            ','P 42            ','P 43            ',
      'I 4             ','I 41            ','P -4            ','I -4            ','P 4/M           ',
      'P 42/M          ','P 4/N           ','P 42/N          ','I 4/M           ','I 41/A          ',
      'P 4 2 2         ','P 4 21 2        ','P 41 2 2        ','P 41 21 2       ','P 42 2 2        ',
      'P 42 21 2       ','P 43 2 2        ','P 43 21 2       ','I 4 2 2         ','I 41 2 2        ',
      'P 4 M M         ','P 4 B M         ','P 42 C M        ','P 42 N M        ','P 4 C C         ',
      'P 4 N C         ','P 42 M C        ','P 42 B C        ','I 4 M M         ','I 4 C M         ',
      'I 41 M D        ','I 41 C D        ','P -4 2 M        ','P -4 2 C        ','P -4 21 M       ',
      'P -4 21 C       ','P -4 M 2        ','P -4 C 2        ','P -4 B 2        ','P -4 N 2        ',
      'I -4 M 2        ','I -4 C 2        ','I -4 2 M        ','I -4 2 D        ',
      'P 4/M M M       ','P 4/M C C       ','P 4/N B M       ','P 4/N N C       ','P 4/M B M       ',
      'P 4/M N C       ','P 4/N M M       ','P 4/N C C       ','P 42/M M C      ','P 42/M C M      ',
      'P 42/N B C      ','P 42/N N M      ','P 42/M B C      ','P 42/M N M      ','P 42/N M C      ',
      'P 42/N C M      ','I 4/M M M       ','I 4/M C M       ','I 41/A M D      ','I 41/A C D      ',
      'P 3             ','P 31            ','P 32            ','R 3             ','P -3            ',
      'R -3            ','P 3 1 2         ','P 3 2 1         ','P 31 1 2        ','P 31 2 1        ',
      'P 32 1 2        ','P 32 2 1        ','R 3 2           ','P 3 M 1         ','P 3 1 M         ',
      'P 3 C 1         ','P 3 1 C         ','R 3 M           ','R 3 C           ','P -3 1 M        ',
      'P -3 1 C        ','P -3 M 1        ','P -3 C 1        ','R -3 M          ','R -3 C          ',
      'P 6             ','P 61            ','P 65            ','P 62            ','P 64            ',
      'P 63            ','P -6            ','P 6/M           ','P 63/M          ','P 6 2 2         ',
      'P 61 2 2        ','P 65 2 2        ','P 62 2 2        ','P 64 2 2        ',
      'P 63 2 2        ','P 6 M M         ','P 6 C C         ','P 63 C M        ','P 63 M C        ',
      'P -6 M 2        ','P -6 C 2        ','P -6 2 M        ','P -6 2 C        ','P 6/M M M       ',
      'P 6/M C C       ','P 63/M C M      ','P 63/M M C      ','P 2 3           ','F 2 3           ',
      'I 2 3           ','P 21 3          ','I 21 3          ','P M 3           ','P N 3           ',
      'F M 3           ','F D 3           ','I M 3           ','P A 3           ','I A 3           ',
      'P 4 3 2         ','P 42 3 2        ','F 4 3 2         ','F 41 3 2        ','I 4 3 2         ',
      'P 43 3 2        ','P 41 3 2        ','I 41 3 2        ','P -4 3 M        ','F -4 3 M        ',
      'I -4 3 M        ','P -4 3 N        ','F -4 3 C        ','I -4 3 D        ','P M 3 M         ',
      'P N 3 N         ','P M 3 N         ','P N 3 M         ','F M 3 M         ','F M 3 C         ',
      'F D 3 M         ','F D 3 C         ','I M 3 M         ','I A 3 D         ','C 1             ',
      'C -1            ']
sg = [string.strip() for string in sg]

# GULP contains at least some errors in the space groups.  Thus we have to somehow hack things until they work.

sgdict = {}

missing_minus = set(range(200, 207))
missing_minus.union(range(221,231))

for i, name in enumerate(sg):
    num = i + 1
    assert name not in sgdict
    sgdict[name] = num
    if num in missing_minus:
        correct_name = name.replace('3', '-3')
        assert correct_name not in sgdict
        sgdict[correct_name] = num

def get_spacegroup_number(name):
    return sgdict[name]

#for i
#print(len(sg))
#from ase.spacegroup import Spacegroup
#asesym = []
#for i in range(1, 231):
#    asesym.append(Spacegroup(i).symbol)

#for i, (s1, s2) in enumerate(zip(asesym, sg)):
#    is_ok = (s1.lower() == s2.lower())
#    print('%3d %10s %10s %10s' % (i + 1, s1, s2, is_ok))




from abc import abstractmethod
from typing import Collection, Tuple, List

from optimade.server.mappers import BaseResourceMapper
from optimade.filterparser import LarkParser
from optimade.models import EntryResource
from optimade.server.query_params import EntryListingQueryParams


class EntryCollection(Collection):  # pylint: disable=inherit-non-class
    def __init__(
        self,
        collection,
        resource_cls: EntryResource,
        resource_mapper: BaseResourceMapper,
    ):
        self.collection = collection
        self.parser = LarkParser()
        self.resource_cls = resource_cls
        self.resource_mapper = resource_mapper

    def __len__(self):
        return self.collection.count()

    def __iter__(self):
        return self.collection.find()

    def __contains__(self, entry):
        return self.collection.count(entry) > 0

    def get_attribute_fields(self) -> set:
        schema = self.resource_cls.schema()
        attributes = schema["properties"]["attributes"]
        if "allOf" in attributes:
            allOf = attributes.pop("allOf")
            for dict_ in allOf:
                attributes.update(dict_)
        if "$ref" in attributes:
            path = attributes["$ref"].split("/")[1:]
            attributes = schema.copy()
            while path:
                next_key = path.pop(0)
                attributes = attributes[next_key]
        return set(attributes["properties"].keys())

    @abstractmethod
    def find(
        self, params: EntryListingQueryParams
    ) -> Tuple[List[EntryResource], int, bool, set]:
        """
        Fetches results and indicates if more data is available.

        Also gives the total number of data available in the absence of page_limit.

        Args:
            params (EntryListingQueryParams): entry listing URL query params

        Returns:
            Tuple[List[Entry], int, bool, set]: (results, data_returned, more_data_available, fields)

        """

    def count(self, **kwargs):
        return self.collection.count(**kwargs)



# Copyright 2016-2018 Ask Hjorth Larsen, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Main author and maintainer: Ask Hjorth Larsen <asklarsen@gmail.com>

from __future__ import print_function
import os
import sys
import re
import logging

import numpy as np
from ase import Atoms
from ase.spacegroup import crystal
#from ase.data import chemical_symbols

from nomadcore.simple_parser import mainFunction, SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.unit_conversion.unit_conversion \
    import register_userdefined_quantity, convert_unit

from .functionals import functionals


parser_info = {'name': 'molcas-parser', 'version': '1.0'}

# OK
#    "program_basis_set_type",
#    "program_name",
#    "program_version",
#    "configuration_periodic_dimensions"
#    "simulation_cell",  # molcas never has a cell



#    "atom_labels",
#    "atom_positions",
#    "energy_total",

#    'section_system',
#    'section_method',
#    'section_frame_sequence',
#    'section_sampling_method',
#    'single_configuration_to_calculation_method_ref',
#    'single_configuration_calculation_to_system_ref',
#    'atom_forces_raw',
#    'frame_sequence_local_frames_ref',
#    'frame_sequence_to_sampling_ref',

#    'XC_functional_name',
#    'smearing_kind',
#    'smearing_width'
#    'eigenvalues_kpoints',
#    'eigenvalues_values',
#    'eigenvalues_occupation',
#    'band_k_points',
#    'band_energies',
#    'band_segm_start_end',
#    'band_segm_labels',
#    'dos_energies',
#    'dos_values',
#    'section_XC_functionals',



class MolcasContext(object):
    def __init__(self):
        self.data = {}
        self.last_line = None
        self.current_module_name = None
        self.section_refs = {}
        self.basis_info = []
        self.buffered_lines = {}

    def onOpen_section_method(self, backend, gindex, section):
        self.section_refs['method'] = gindex

    def onOpen_section_system(self, backend, gindex, section):
        self.section_refs['system'] = gindex

    def startedParsing(self, fname, parser):
        pass

    def adhoc_store_line(self, parser):
        self.last_line = parser.fIn.readline()

    def adhoc_pushback_last(self, parser):
        assert self.last_line is not None
        parser.fIn.pushbackLine(self.last_line)
        self.last_line = None

    def onClose_section_single_configuration_calculation(self, backend, gindex, section):
        backend.addValue('single_configuration_to_calculation_method_ref',
                         self.section_refs['method'])
        backend.addValue('single_configuration_calculation_to_system_ref',
                         self.section_refs['system'])

        #print(self.buffered_lines.keys())

        def lines2arrays(name):
            elines = self.buffered_lines.pop('energy_%s' % name)
            olines = self.buffered_lines.pop('occupations_%s' % name)
            energies = []
            occupations = []
            assert len(elines) == len(olines)
            for l1, l2 in zip(elines, olines):
                energies += l1.split()[1:]  # Energy <numbers>
                occupations += l2.split()[2:]  # Occ. No.  <numbers>
            assert len(energies) == len(occupations)
            energies = np.array(energies).astype(float)
            energies = convert_unit(energies, 'hartree')
            occupations = np.array(occupations).astype(float)

            # Sort by energy:
            args = np.argsort(energies)
            energies = energies[args]
            occupations = occupations[args]
            return energies.reshape(1, 1, -1), occupations.reshape(1, 1, -1)

        energies = None

        try:
            energies, occupations = lines2arrays('nospin')
            nspins = 1
            nstates = energies.shape[-1]
        except KeyError:
            pass

        try:
            ealpha, oalpha = lines2arrays('alpha')
            ebeta, obeta = lines2arrays('beta')
        except KeyError:
            pass
        else:
            # The spinup and spindown (alpha, beta)
            # arrays can have different lengths
            # so we must allocate arrays that are big enough
            # and fill out with zeros or something else
            assert energies is None
            nspins = 2
            nstates = max(ealpha.shape[-1], ebeta.shape[-1])
            energies = np.empty((nspins, 1, nstates))
            # We will only fill out the values that we have.
            # The rest will be nan:
            energies.fill(np.nan)
            occupations = np.zeros((nspins, 1, nstates))
            energies[0, :, :ealpha.shape[2]] = ealpha[0]
            occupations[0, :, :ealpha.shape[2]] = oalpha[0]
            energies[1, 0, :ebeta.shape[2]] = ebeta[0]
            occupations[1, 0, :ebeta.shape[2]] = obeta[0]

        if energies is not None:
            for arr in [energies, occupations]:
                assert arr.shape == (nspins, 1, nstates)
            g = backend.openSection('section_eigenvalues')
            backend.addArrayValues('eigenvalues_values', energies)
            backend.addArrayValues('eigenvalues_occupation', occupations)
            backend.closeSection('section_eigenvalues', g)


    def onClose_section_method(self, backend, gindex, section):

        for line in self.basis_info:
            m = re.match('\s*Basis set label:(.*?)\.(.*?)\s*$', line)
            atomlabel, name = m.group(1, 2)
            # The terrible syntax in molcas has a lot of dots everywhere.
            # Replace sequences of dots with single dots to make things
            # reasonable.
            name = re.sub(r'\.+', '.', name).strip('.')
            g = backend.openSection('x_molcas_section_basis')
            backend.addValue('x_molcas_basis_atom_label', atomlabel)
            backend.addValue('x_molcas_basis_name', name)
            backend.closeSection('x_molcas_section_basis', g)
        del self.basis_info[:]

        #print('SECTION', section)
        #methods = section['x_molcas_method']
        xc = []

        methods = section['x_molcas_method_name']
        assert len(methods) == 1, methods
        method = methods[0]

        uhf = method.startswith('UHF ')
        if uhf:
            method = method[4:].strip()
        backend.addValue('x_molcas_uhf', uhf)

        if method == 'SCF':
            # XXX distinguish somehow between these two?
            esm = 'DFT'
            xc = ['HF_X']
        elif method in ['RASSCF', 'RAS-CI', 'CASPT2', 'CCSDT']:
            esm = method
        elif method in ['CASVB', 'MRCI']:
            esm = 'MRCISD'
        elif method == 'MCPF':
            esm = method  ## Not defined by nomad
        elif method == 'MBPT2':
            esm = 'MP2'
        elif method in functionals:
            esm = 'DFT'
            xc = functionals[method]
            assert xc is not None, 'no xc: %s' % method
        else:
            raise ValueError('method: %s' % method)

        backend.addValue('electronic_structure_method', esm)
        for xcfunc in xc:
            g = backend.openSection('section_XC_functionals')
            backend.addValue('XC_functional_name', xcfunc)
            backend.closeSection('section_XC_functionals', g)


        if 0:
            if methods is not None:
                assert len(methods) == 1
                method = methods[0]
                scfnames = section['x_molcas_method_name']
                assert len(scfnames) == 1, scfnames
                scfname = scfnames[0]
                if scfname == 'SCF':
                    backend.addValue('electronic_structure_method', 'DFT')
                    g = backend.openSection('section_XC_functionals')
                    backend.addValue('XC_functional_name', 'HF_X')
                    backend.closeSection('section_XC_functionals', g)

        #print(self.current_module_name)
        #scftype = section['x_molcas_method_name']
        #if scftype is not None:
        #    print('SCF', scftype)

    def onClose_section_system(self, backend, gindex, section):
        matrix = self.data.pop('coordinates')
        assert matrix.shape[1] == 4
        atom_labels = matrix[:, 0]
        coords = matrix[:, 1:4].astype(float)
        assert coords.shape[1] == 3
        coords = convert_unit(coords, 'bohr')

        backend.addArrayValues('atom_labels', atom_labels)
        backend.addArrayValues('atom_positions', coords)
        backend.addArrayValues('configuration_periodic_dimensions', np.zeros(3, bool))

    def adhoc_add_basis_set(self, parser):
        line = parser.fIn.readline()
        self.basis_info.append(line)

    def adhoc_add_energies(self, parser):
        line = parser.fIn.readline()
        self.energies.append(line)

    def adhoc_add_occupations(self, parser):
        line = parser.fIn.readline()
        self.occupations.append(line)

    # multi_sm copied from Siesta/GULP
    def multi_sm(self, name, startpattern, linepattern, endmatcher=None,
                 conflict='fail',  # 'fail', 'keep', 'overwrite'
                 *args, **kwargs):

        pat = re.compile(linepattern)  # XXX how to get compiled pattern?
        ngroups = pat.groups

        allgroups = []
        def addline(parser):
            line = parser.fIn.readline()
            match = pat.match(line)
            assert match is not None
            thislinegroups = match.groups()
            assert len(thislinegroups) == ngroups
            allgroups.append(thislinegroups)

        def savearray(parser):
            arr = np.array(allgroups, dtype=object)
            del allgroups[:]
            if name in self.data:
                if conflict == 'fail':
                    raise ValueError('grrr %s %s' % (name, self.data[name]))
                elif conflict == 'keep':
                    return  # Do not save array
                elif conflict == 'overwrite':
                    pass
                else:
                    raise ValueError('Unknown keyword %s' % conflict)
            #if arr.size > 0:
            if arr.size == 0:
                arr.shape = (0, ngroups)
            self.data[name] = arr

        if endmatcher is None:
            endmatcher = r'.*'

        if hasattr(endmatcher, 'swapcase'):
            endmatcher = SM(endmatcher,
                            endReStr='',
                            forwardMatch=True,
                            name='%s-end' % name,
                            adHoc=savearray)

        sm = SM(startpattern,
                name=name,
                subMatchers=[
                    SM(linepattern,
                       name='%s-line' % name,
                       repeats=True,
                       forwardMatch=True,
                       required=True,
                       adHoc=addline),
                    endmatcher,
                ], **kwargs)
        return sm

context = MolcasContext()

#def get_inputfile_echo_sm():
#    m = SM(r'\+\+\s*----+\s*Input file\s*---+',
#           endReStr='--\s*-------+',
#           sections=['section_method'],
#           name='inputfile')
#    return m


def get_system_sm():
    m = SM(r'\s*Molecular structure info:',
           name='structure',
           sections=['section_system'],
           subMatchers=[
               SM(r'\s*--------+', name='bar'),
               SM(r'.*?Cartesian Coordinates / Bohr, Angstrom\s*\*+',
                  name='coords',
                  subMatchers=[
                      context.multi_sm('coordinates',
                                       r'\s*Center\s*Label\s*x\s*y\s*z\s*x\s*y\s*z',
                                       r'\s*\d+\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)')
                  ])
           ])
    return m


def get_finalresults_sm():  # Other modes than SCF/KS-DFT?
    m = SM(r'\s*\*\s*SCF/KS-DFT Program, Final results',
           endReStr='\s*Molecular Properties',  # No clue when to really stop matching this
           sections=['section_single_configuration_calculation'],
           name='results',
           subMatchers=[
               SM(r'\s*Total SCF energy\s*(?P<energy_total__hartree>\S+)', name='etot')
           ])
    return m

def get_header_sm():
    m = SM(r'[\s^]*M O L C A S',
           endReStr=r'\s*For the author list and the recommended.*',
           name='header',
           subMatchers=[
               SM(r'[\s^]*version\s*(?P<program_version>.*?)\s*$',
                  name='version')
           ])
    return m

# Does not include 'auto'
molcas_modules = ('alaska|caspt2|casvb|ccsdt|cpf|expbas|ffpt|gateway|'
                  'genano|grid_it|guessorb|guga|last_energy|loprop|mbpt2|'
                  'mckinley|mclr|motra|mrci|numerical_gradient|rasscf|'
                  'rassi|scf|seward|slapaf|vibrot').split('|')

ignore_modules = {'alaska',  # OK
                  'numerical_gradient',  # OK
                  'rassi',  # OK
                  'motra',
                  'guga',
                  'ffpt',
                  'genano',
                  'grid_it',
                  'loprop',
                  'guessorb',
                  'expbas',
                  'vibrot',
                  'mckinley'
                  #'last_energy',
}

def get_anymodule_sms():
    sms = []
    for modulename in molcas_modules:
        adhoc = None
        if modulename not in ignore_modules:
            class AdHoc:
                def __init__(self, name):
                    self.name = name
                def __call__(self, parser):
                    raise ValueError('module: %s' % self.name)
            adhoc = AdHoc(modulename)
        sms.append(SM(r'--- Start Module:\s*%s' % modulename,
                      adHoc=adhoc,
                      name=modulename))
    return sms

#def module_sm(name, *args, **kwargs):
#    m = SM(r'--- Start Module:\s*(%s)' % name, *args, name=name, **kwargs)
#    return m

# mw = "module-wrapper" for SM
# The idea is to guarantee that matchers modules that do not always write a proper header (e.g. SLAPAF)
# are invoked when the module starts, but that the matcher's sections are only opened if recognizable
# output is written (sometimes they write nothing).
#
# Typically, pass a matcher which opens some sections but will only match actual output of the module
# This wrapper will call the wrapped matcher when the module is started, but the wrapped matcher
# will only match if the output is actually there, and also will only open its sections in that case.
def mw(modname, *args, **kwargs):
    m = SM(r'--- Start Module:\s*%s' % modname,
           name=modname,
           subMatchers=[
               SM(*args, **kwargs)
           ])
    return m

def mod_pattern(lower, upper=None):
    if upper is None:
        upper = lower.upper()
    return (r'---\s*Start Module:\s*{0}|'
            r'\s*MOLCAS executing module\s*{1}'.format(lower, upper))

# Define these separately or not?
def startmod_pattern(name):
    return r'---\s*Start Module:\s*{}'.format(name)
def execmod_pattern(name):
    return r'\s*MOLCAS executing module\s*{}'.format(name)

def gateway_seward():
    m = SM(mod_pattern(r'(:?gateway|seward)'),
           #weak=True,
           name='gate/seward',
           #r'\s*MOLCAS executing module (GATEWAY|SEWARD)', name='seward',
           subMatchers=[
               SM(r'\s*Basis set label:(.*?)\s*$',
                  forwardMatch=True,
                  repeats=True,
                  adHoc=context.adhoc_add_basis_set,
                  name='basis'),
               get_system_sm()
           ])
    return m

def eigenvalue_line_matcher(pattern, spinname):

    class AdHocLineAccumulator:
        def __init__(self, var):
            self.var = var

        def __call__(self, parser):
            line = parser.fIn.readline()
            context.buffered_lines.setdefault(self.var, []).append(line)
            #print(context.buffered_lines)
            #sdfkjsdfkj

    m = SM(pattern, name=spinname,
           subMatchers=[
               SM(r'\s*Orbital',
                  repeats=True,
                  subMatchers=[
                      SM(r'\s*Energy\s*(.*)', name='energies', adHoc=AdHocLineAccumulator('energy_%s' % spinname),
                         forwardMatch=True),
                      SM(r'\s*Occ\. No\.\s*(.*)', name='occ', adHoc=AdHocLineAccumulator('occupations_%s' % spinname),
                         forwardMatch=True)
                  ])
           ])
    return m

def scf_rasscf():
    m = SM(mod_pattern(r'(:?ras)?scf'),
           name='(ras)scf', #r'(?:ras)?scf', r'\s*MOLCAS executing module (?:SCF|RASSCF)', name='scf',
           #sections=['section_method', 'section_single_configuration_calculation'],
           subMatchers=[  # XXX opens method section even if there is no outpout
               SM(r'\s*(?P<x_molcas_method_name>.*?)\s*iterations: Energy and convergence statistics',
                  sections=['section_method'],
                  name='scfiter'),
               SM(r'\s*CI only, no orbital optimization',
                  sections=['section_method'],
                  fixedStartValues={'x_molcas_method_name': 'RAS-CI'}),
               SM(r'\s*\*\s*Final Results\s*\*',
                  name='finalresults',
                  sections=['section_single_configuration_calculation'],
                  subMatchers=[
                      SM(r'\s*Total \S+ energy\s*(?P<energy_total__hartree>\S+)', name='E-tot'),
                      SM(r'\s*RASSCF root number\s*\d+\s*Total energy\s*=\s*(?P<energy_total__hartree>\S+)', name='E-ras'),
                      SM(r'\s*Molecular orbitals:',
                         name='mol-orb',
                         endReStr=r'\s*Molecular Charges:',
                         subMatchers=[
                             eigenvalue_line_matcher(r'\s*Title:\s*.*?\s*orbitals\s*$', 'nospin'),
                             eigenvalue_line_matcher(r'\s*Title:\s*.*?\s*orbitals\s*\(alpha\)', 'alpha'),
                             eigenvalue_line_matcher(r'\s*Title:\s*.*?\s*orbitals\s*\(beta\)', 'beta')
                         ])
                  ])
           ])
    return m

def caspt2():
    m = SM(mod_pattern(r'caspt2'),
           #caspt2', r'\s*MOLCAS executing module CASPT2',
           name='caspt2',  # XXX opens section even if there is no output
           sections=['section_method', 'section_single_configuration_calculation'],
           fixedStartValues={'x_molcas_method_name': 'CASPT2'},
           subMatchers=[
               SM(r'\s*Total energy:\s*(?P<energy_total__hartree>\S+)', name='E-caspt2')
           ])
    return m

def molcas_main_loop_sm():
    m = SM(r'--- Start Module:\s*(?:%s)' % '|'.join(molcas_modules),
           endReStr='--- Stop Module:',
           repeats=True,
           forwardMatch=True,
           name='module',
           subMatchers=[
               gateway_seward(),
               scf_rasscf(),
               caspt2(),
               SM(r'%s|%s' % (mod_pattern(r'slapaf'),
                              # Sometimes the damn thing starts without a proper header!!  ARGH!
                              r's*Specification of the internal coordinates according to the user-defined'),
                  name='slapaf',
                  # The slapaf module writes forces, but only for a symmetry-reduced set of atoms.
                  # Therefore they are not useful to parse.  We could parse the set of symmetries probably
                  # and somehow make it possible to get the full set of Cartesian forces,
                  # but the operation would be very complicated.
                  #
                  # Let us consider parsing the reduced forces and
                  # symmetries the day that someone wants to implement
                  # this symmetry-unreduction (this will not happen).
                  subMatchers=[
                      SM(r'\*\s*Energy Statistics for Geometry Optimization',
                         name='opt-loop',
                         endReStr=r'\s*$',
                         sections=['section_single_configuration_calculation'],
                         subMatchers=[
                             SM(r'\s*\d+\s*\S+\s*', name='opt-iter', repeats=True, forwardMatch=True, adHoc=context.adhoc_store_line),
                             SM(r'\s*$', forwardMatch=True, adHoc=context.adhoc_pushback_last, name='pushback'),
                             # Molcas is supposed to have spaces between the numbers, but with grdNorm/Max,
                             # it sometimes writes 0.051834-0.037589 in one.  Thus we use -?[^\s-]+ to match such a number.
                             SM(r'\s*\d+\s*(?P<energy_total__hartree>\S+)\s*'
                                r'\S+\s*(?P<x_molcas_slapaf_grad_norm>-?[^\s-]+)\s*'
                                r'(?P<x_molcas_slapaf_grad_max>\S+)',
                                name='opt-iter', repeats=True),
                         ]),
                      SM(r'\s*\*\s*Nuclear coordinates for the next iteration / Bohr',
                         endReStr=r'\s*$',
                         name='newcoords',
                         sections=['section_system'],
                         subMatchers=[
                             context.multi_sm('coordinates',
                                              r'\s*ATOM\s*X\s*Y\s*Z',
                                              r'\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)')
                             ])
                  ]),
               # last_energy just means it calls some other modules with annoyingly irregular output format
               SM(mod_pattern(r'last_energy'),
                  name='last',
                  subMatchers=[
                      gateway_seward(),
                      scf_rasscf(),
                      caspt2(),
                  ]),
               SM(mod_pattern('casvb'),
                  name='casvb',
                  sections=['section_method'],
                  fixedStartValues={'x_molcas_method_name': 'CASVB'},
                  subMatchers=[
                      SM('\s*CASSCF energy\s*:\s*(?P<energy_total__hartree>\S+)',
                         sections=['section_single_configuration_calculation'],
                         name='e-tot')
                  ]),
               SM(mod_pattern('mrci'),
                  name='mrci',
                  sections=['section_method'],
                  fixedStartValues={'x_molcas_method_name': 'MRCI'},
                  subMatchers=[
                      SM(r'\s*CI ENERGY:\s*(?P<energy_total__hartree>\S+)',
                         sections=['section_single_configuration_calculation'])
                  ]),
               SM(mod_pattern('cpf', 'CPFMCPF'),
                  name='cpf',
                  sections=['section_method'],
                  fixedStartValues={'x_molcas_method_name': 'MCPF'},
                  subMatchers=[
                      SM(r'\s*FINAL MCPF ENERGY\s*(?P<energy_total__hartree>\S+)',
                         sections=['section_single_configuration_calculation'])
                  ]),
               SM(mod_pattern('ccsdt', r'CCSD\(T\)'),
                  name='ccsdt',
                  sections=['section_method'],
                  fixedStartValues={'x_molcas_method_name': 'CCSDT'},
                  subMatchers=[
                      SM(r'\s*Total energy \(diff\)\s*:\s*(?P<energy_total__hartree>\S+)',
                         sections=['section_single_configuration_calculation'])
                  ]),
               SM(mod_pattern('mbpt2'),
                  name='mbpt2',
                  fixedStartValues={'x_molcas_method_name': 'MBPT2'},
                  sections=['section_method', 'section_single_configuration_calculation'],
                  subMatchers=[
                      SM(r'\s*Total energy\s*=\s*(?P<energy_total__hartree>\S+)\s*a\.u\.')
                  ]),
               SM(mod_pattern('mclr'),
                  name='mclr',
                  subMatchers=[
                      SM(r'\s*\*\s*Harmonic frequencies in cm-1\s*\*',
                         endReStr=r'\s*\*\s*THERMOCHEMISTRY',
                         sections=['section_single_configuration_calculation'],
                         subMatchers=[
                             SM(r'\s*Symmetry\s*(?P<x_molcas_frequency_symmetry>\S+)',
                                name='symm',
                                repeats=True,
                                sections=['x_molcas_section_frequency'],
                                subMatchers=[
                                    SM(r'\s*Freq\.\s*i(?P<x_molcas_imaginary_frequency_value__cm_1>\S+)', name='ifreq'),
                                    SM(r'\s*Freq\.\s*(?P<x_molcas_frequency_value>\S+)', name='freq'),
                                    SM(r'\s*Intensity:\s*(?P<x_molcas_frequency_intensity__km_mol_1>\S+)', name='intensity')
                                ])
                         ]),
                  ]),
           ] + get_anymodule_sms())
    return m

def main_file_description():
    return SM(
        name='root',
        weak=True,
        startReStr='',
        fixedStartValues={'program_name': 'MOLCAS',
                        'program_basis_set_type': 'gaussians'},
        sections=['section_run'],
        subMatchers=[
            get_header_sm(),
            #get_inputfile_echo_sm(),
            molcas_main_loop_sm(),
            SM(r'x^',  # force parser to parse the whole file
            name='impossible')
        ])


class MolcasParser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        global context
        context = MolcasContext()

        from unittest.mock import patch
        logging.debug('molcas parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("molcas.nomadmetainfo.json")
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                mainFileDescription=main_file_description(),
                metaInfoEnv=None,
                parserInfo=parser_info,
                cachingLevelForMetaName={},
                superContext=context,
                superBackend=backend)

        return backend



from nomad.metainfo import MSection, Section, SubSection, Quantity
from nomad.metainfo.mongoengine_extension import Mongo, MongoDocument


class DOSSimilarity(MSection):
    m_def = Section(
        a_mongo=MongoDocument()
    )
    material_ids = Quantity(
        type=str,
        shape=["n_similar_materials"],
        a_mongo=Mongo(),
    )
    values = Quantity(
        type=float,
        shape=["n_similar_materials"],
        a_mongo=Mongo(),
    )


class Similarity(MSection):
    m_def = Section(
        a_mongo=MongoDocument()
    )
    electronic_dos = SubSection(sub_section=DOSSimilarity.m_def, repeats=False)


class Material(MSection):
    m_def = Section(
        a_mongo=MongoDocument()
    )
    material_id = Quantity(
        type=str,
        a_mongo=Mongo(primary_key=True)
    )
    similarity = SubSection(sub_section=Similarity.m_def, repeats=False)



# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from phonopyparser.phonopy_calculators import PhonopyCalculatorInterface

from nomad.parsing.parser import FairdiParser
from .metainfo import m_env


class PhonopyParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/phonopy', code_name='Phonopy', code_homepage='https://phonopy.github.io/phonopy/',
            mainfile_name_re=(r'(.*/phonopy-FHI-aims-displacement-0*1/control.in$)|(.*/phonon.yaml)')
        )

    def parse(self, filepath, archive, logger=None):
        self._metainfo_env = m_env

        interface = PhonopyCalculatorInterface(filepath, archive, logger)
        interface.parse()



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import unittest

class TestAtomStyles(unittest.TestCase):

    def test_style_atomic(self):
        style_atom = atom_style('atomic')
        self.assertIsInstance(style_atom, atom_style)

    def test_style_bond(self):
        style_bond = atom_style('bond')
        self.assertIsInstance(style_bond, atom_style)

    def test_style_full(self):
        style_full = atom_style('full')
        self.assertIsInstance(style_full, atom_style)

    def test_style_body(self):
        style_body = atom_style('body', 'nparticle', 2, 10)
        self.assertIsInstance(style_body, atom_style)

    def test_style_hybrid(self):
        style_hybrid = atom_style('hybrid', 'charge', 'bond')
        self.assertIsInstance(style_hybrid, atom_style)

    def test_style_hybrid(self):
        style_hybrid = atom_style('hybrid', 'charge', 'body', 'nparticle', 2, 5)
        self.assertIsInstance(style_hybrid, atom_style)

    def test_style_template(self):
        style_template = atom_style('template', 'myMols')
        self.assertIsInstance(style_template, atom_style)




# Copyright 2015-2018 Franz Knuth, Fawzi Mohamed, Wael Chibani, Ankit Kariryaa, Lauri Himanen, Danio Brambila
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import logging
import sys
from nomadcore.baseclasses import ParserInterface, MainHierarchicalParser

from fhiaimsparser.FhiAimsParser import FhiAimsParserContext, build_FhiAimsMainFileSimpleMatcher, get_cachingLevelForMetaName, getParserInfo


class FHIaimsParser(ParserInterface):
    """This class provides an object-oriented access to the parser.
    """
    def __init__(
            self, metainfo_to_keep=None, backend=None,
            default_units=None, metainfo_units=None, debug=False,
            log_level=logging.ERROR, store=True):
        super(FHIaimsParser, self).__init__(
            metainfo_to_keep, backend, default_units,
            metainfo_units, debug, log_level, store)

    def setup_version(self):
        """Setups the version by possbily investigating the output file and the
        version specified in it.
        """
        # Setup the root folder to the fileservice that is used to access files
        dirpath, filename = os.path.split(self.parser_context.main_file)
        dirpath = os.path.abspath(dirpath)
        self.parser_context.file_service.setup_root_folder(dirpath)
        self.parser_context.file_service.set_file_id(filename, "output")

        # Setup the correct main parser possibly based on the version
        self.main_parser = FHIaimsMainParser(self.parser_context)

    @staticmethod
    def get_mainfile_regex():
        regex_str = (
            "\s*Invoking FHI-aims \.\.\.\n"
            "\s*Version "
        )
        return regex_str

    def get_metainfo_filename(self):
        return "fhi_aims.nomadmetainfo.json"

    def get_parser_info(self):
        return getParserInfo()


fhiAimsMainFileSimpleMatcher = build_FhiAimsMainFileSimpleMatcher()


class FHIaimsMainParser(MainHierarchicalParser):
    """The main parser class that is called for all run types.
    """
    def __init__(self, parser_context):
        """
        """
        super(FHIaimsMainParser, self).__init__(parser_context)
        self.root_matcher = fhiAimsMainFileSimpleMatcher
        self.caching_levels = get_cachingLevelForMetaName(
            parser_context.metainfo_env)
        self.super_context = FhiAimsParserContext()


def main():
    parser = FHIaimsParser()
    results = parser.parse(sys.argv[1])
    return results


if __name__ == "__main__":
    main()
